import math
from copy import deepcopy

from magnumapi.geometry.Geometry import Geometry
from magnumapi.geometry.SlottedGeometry import SlottedRelativeCosThetaGeometry
from magnumapi.geometry.blocks.Block import Block
from magnumapi.geometry.blocks.CosThetaBlock import AbsoluteCosThetaBlock
from magnumapi.geometry.primitives.Line import Line


class GeometryChange:

    @classmethod
    def limit_minimum_phi_r(cls, geometry, min_length=1):
        geometry = deepcopy(geometry)
        for layer_def in geometry.layer_defs:
            for blocks_index, block_index in enumerate(layer_def.blocks):
                if blocks_index > 0:
                    index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                    radius = geometry.blocks[index_in_blocks].block_def.radius
                    phi_r = geometry.blocks[index_in_blocks].block_def.phi_r
                    if radius > 0:
                        geometry.blocks[index_in_blocks].block_def.phi_r = max(phi_r, math.degrees(min_length / radius))

        return geometry


    @classmethod
    def update_spar_thickness(cls, geometry: SlottedRelativeCosThetaGeometry, decoded):
        geometry = deepcopy(geometry)

        for layer_index, layer_def in enumerate(geometry.layer_defs):
            layer_def.spar_thickness = decoded.get('spar_thickness:%d:' % (layer_index + 1), layer_def.spar_thickness)

        # ToDo: here one should use a builder!!!
        return SlottedRelativeCosThetaGeometry(r_aperture=geometry.r_aperture,
                                               blocks=geometry.blocks,
                                               layer_defs=geometry.layer_defs)

    @classmethod
    def update_nco_varying_blocks(cls, geometry, decoded):
        geometry = deepcopy(geometry)
        block_index = 1
        blocks = []
        # for each layer
        for layer_index, layer_def in enumerate(geometry.layer_defs):
            # find a reference block - first block in layer_def
            block_layer_index_ref = layer_def.blocks[0]
            block_index_ref = geometry.get_index_in_blocks_for_layer_block_index(block_layer_index_ref)

            # find all ncos that belong to that layer and are non-zero
            nco_values = {key: value for key, value in decoded.items() if value and 'nco:%d' % (layer_index + 1) in key}

            # update a copy of the reference block as long as there are non-zero blocks - todo: sort the keys
            layer_blocks = []
            for key, value in nco_values.items():
                block_def_ref = deepcopy(geometry.blocks[block_index_ref])
                block_def_ref.block_def.nco = value
                block_def_ref.block_def.no = block_index
                layer_blocks.append(block_index)
                block_index += 1
                blocks.append(block_def_ref)

            layer_def.blocks = layer_blocks

        return SlottedRelativeCosThetaGeometry(r_aperture=geometry.r_aperture,
                                               blocks=blocks,
                                               layer_defs=geometry.layer_defs)

    @classmethod
    def update_alpha_radial(cls, geometry, chromosome):
        geometry = deepcopy(geometry)
        for block_index_variable, value in chromosome.items():
            if 'alpha_rad_r' in block_index_variable:
                variable, layer_index, block_layer_index = block_index_variable.split(':')
                layer_index = int(layer_index)
                block_layer_index = int(block_layer_index)

                # block index
                if block_layer_index <= len(geometry.layer_defs[layer_index - 1].blocks):
                    no_block = geometry.layer_defs[layer_index - 1].blocks[block_layer_index - 1]
                    block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
                    geometry.blocks[block_index].block_def.alpha_r += value

        return geometry

    @classmethod
    def update_phi_r(cls, geometry, chromosome):
        geometry = deepcopy(geometry)
        for block_index_variable, value in chromosome.items():
            if 'phi_r' in block_index_variable:
                variable, layer_index, block_layer_index = block_index_variable.split(':')
                layer_index = int(layer_index)
                block_layer_index = int(block_layer_index)

                # block index
                if block_layer_index <= len(geometry.layer_defs[layer_index - 1].blocks):
                    no_block = geometry.layer_defs[layer_index - 1].blocks[block_layer_index - 1]
                    block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
                    geometry.blocks[block_index].block_def.phi_r = value

        return geometry

    @classmethod
    def calculate_radial_alpha(cls, geometry: Geometry) -> Geometry:
        """ Method correcting radiality of cos-theta blocks in a geometry.

        """
        # check whether areas were initialize before performing radiality correction
        is_any_area_initialized = any([area for block in geometry.blocks for area in block.areas])

        # Set alpha equal to phi for all blocks except the mid-plane
        geometry = cls._set_alpha_equal_to_phi(geometry)

        # Build blocks and calculate turn positions
        geometry.build_blocks()

        # Calculate alpha corrections
        alpha_corrections = cls._calculate_alpha_corrections(geometry)

        # Correct alpha
        geometry = cls._apply_alpha_correction(geometry, alpha_corrections)

        # Return to the external context of the geometry before entering
        if is_any_area_initialized:
            geometry.build_blocks()
        else:
            geometry.empty_block_areas()

        return geometry

    @staticmethod
    def _set_alpha_equal_to_phi(geometry):
        for layer_def in geometry.layer_defs:
            for block_index in layer_def.blocks[1:]:
                index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                block = geometry.blocks[index_in_blocks]
                if isinstance(block, AbsoluteCosThetaBlock):
                    block.block_def.alpha = block.block_def.phi
        return geometry

    @classmethod
    def _calculate_alpha_corrections(cls, geometry):
        alpha_corrections = []
        for layer_def in geometry.layer_defs:
            for index, block_index in enumerate(layer_def.blocks):
                index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
                block = geometry.blocks[index_in_blocks]
                # Calculate correction for all blocks except the mid-plane one
                if index and isinstance(block, AbsoluteCosThetaBlock):
                    alpha_correction = cls.compute_radial_alpha(block)
                else:
                    alpha_correction = 0.0
                alpha_corrections.append(alpha_correction)

        return alpha_corrections

    @staticmethod
    def compute_radial_alpha(block: Block) -> float:
        """ Static method calculating a radiality correction angle for a cos-theta block. It is based on an assumption
        that block alpha angle (inclination angle) is equal to phi angle (positioning angle).

        :param block: a block instance with initialized areas
        :return: an alpha angle correction factor
        """
        if not block.areas:
            raise IndexError('The list of areas is empty, please build block first with build_block() method.')

        l_down = block.areas[0].get_line(0)
        p_down = l_down.p1
        l_up = block.areas[-1].get_line(2)
        p_up = l_up.p2

        angle_ref = (p_up.get_phi() + p_down.get_phi()) / 2

        alpha_down = angle_ref - Line.calculate_relative_alpha_angle(l_down)
        alpha_up = Line.calculate_relative_alpha_angle(l_up) - angle_ref

        return (alpha_down - alpha_up) / 2

    @staticmethod
    def _apply_alpha_correction(geometry, alpha_corrections):
        for index, block in enumerate(geometry.blocks):
            if isinstance(block, AbsoluteCosThetaBlock):
                block.block_def.alpha += alpha_corrections[index]

        return geometry

    @classmethod
    def update_layer_indexing(cls, geometry) -> Geometry:
        """ Method updating block and layer indexing in case a block has no turns.
        The algorithm works as follows:
        1. Take old block indices and map them to None
        2. Keep blocks with at least one conductor
        3. Update the numbering of blocks; update mapping from old to new indices from point 1
        4. Update block indices in each layer definition with map; if a new index is None, then it means that a block
        was removed

        :return: method returns a dictionary from old to new block indices
        """
        # # initialize map
        geometry = deepcopy(geometry)
        old_to_new_index = {block.block_def.no: None for block in geometry.blocks}

        # # remove the missing blocks
        # # keep only block indices that are not missing
        blocks = []
        for index, block in enumerate(geometry.blocks):
            if block.block_def.nco > 0:
                blocks.append(block)
                old_to_new_index[block.block_def.no] = block.block_def.no

        geometry.blocks = blocks

        # # update layers definition
        for layer_def in geometry.layer_defs:
            layer_def.blocks = [old_to_new_index[block] for block in layer_def.blocks]
            layer_def.blocks = [block for block in layer_def.blocks if block is not None]

        return geometry

    @classmethod
    def update_nco_r(cls, geometry, chromosome):
        geometry = deepcopy(geometry)
        for block_index_variable, value in chromosome.items():
            if 'nco_r' in block_index_variable:
                variable, layer_index, block_layer_index = block_index_variable.split(':')
                layer_index = int(layer_index)
                block_layer_index = int(block_layer_index)

                # ToDo: this requires a correction in case the next one would receive some
                no_block = geometry.layer_defs[layer_index - 1].blocks[block_layer_index - 1]
                block_index = geometry.get_index_in_blocks_for_layer_block_index(no_block)
                block_def_curr = geometry.blocks[block_index].block_def
                block_def_next = geometry.blocks[block_index + 1].block_def
                if value >= 0:
                    block_def_curr.nco += min(value, block_def_next.nco)
                    block_def_next.nco -= min(value, block_def_next.nco)
                else:
                    block_def_curr.nco -= min(abs(value), block_def_curr.nco)
                    block_def_next.nco += min(abs(value), block_def_next.nco)
        geometry = GeometryChange.update_layer_indexing(geometry)
        return geometry



import inspect
from typing import Type

from magnumapi.tool_adapters.ToolAdapter import ToolAdapter
from magnumapi.tool_adapters.roxie.RoxieToolAdapter import TerminalRoxieToolAdapter
from magnumapi.tool_adapters.roxie.RoxieToolAdapter import DockerTerminalRoxieToolAdapter
from magnumapi.tool_adapters.ansys.AnsysToolAdapter import TerminalAnsysToolAdapter
from magnumapi.tool_adapters.ansys.AnsysToolAdapter import MapdlAnsysToolAdapter
from magnumapi.tool_adapters.ansys.AnsysToolAdapter import DockerMapdlAnsysToolAdapter
import magnumapi.commons.json_file as json_file


class ToolAdapterFactory:
    """ A ToolAdapterFactory class producing tool adapter objects from a configuration file either as a json or as a
    dictionary

    """

    @classmethod
    def init_with_json(cls, config_file_path: str) -> Type[ToolAdapter]:
        """ Class method initializing a tool adapter from a json file

        :param config_file_path: an absolute path to a config json file
        :return: an initialized ToolAdapter instance
        """
        config_dct = json_file.read(config_file_path)
        return cls.init_with_dict(config_dct)

    @staticmethod
    def init_with_dict(config_dct: dict) -> Type[ToolAdapter]:
        """ Static method initializing a tool adapter from a dictionary

        :param config_dct: a dictionary with a ToolAdapter configuration parameters
        :return: an initialized ToolAdapter instance
        """
        if config_dct.keys() == inspect.signature(TerminalRoxieToolAdapter).parameters.keys():
            return TerminalRoxieToolAdapter(**config_dct)
        elif config_dct.keys() == inspect.signature(DockerTerminalRoxieToolAdapter).parameters.keys():
            return DockerTerminalRoxieToolAdapter(**config_dct)
        elif config_dct.keys() == inspect.signature(TerminalAnsysToolAdapter).parameters.keys():
            return TerminalAnsysToolAdapter(**config_dct)
        elif config_dct.keys() == inspect.signature(MapdlAnsysToolAdapter).parameters.keys():
            return MapdlAnsysToolAdapter(**config_dct)
        elif config_dct.keys() == inspect.signature(DockerMapdlAnsysToolAdapter).parameters.keys():
            return DockerMapdlAnsysToolAdapter(**config_dct)
        else:
            raise KeyError('The input config definition keys {} do not match any tool adapter constructor signature!'
                           .format(config_dct.keys()))



from unittest import TestCase
from unittest.mock import patch

from magnumapi.tool_adapters.roxie.RoxieToolAdapter import RoxieToolAdapter
from tests.resource_files import create_resources_file_path


class TestRoxieToolAdapter(TestCase):
    @patch("plotly.graph_objects.Figure.show")
    def test_parse_roxie_xml_11T(self, mock_show=None):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/11T/reference/roxieData11T.xml')
        strand_data = RoxieToolAdapter.parse_roxie_xml(roxie_data_xml_path)

        # act
        RoxieToolAdapter.plotly_results(strand_data)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    def test_parse_roxie_xml_16T_formatted(self):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/16T/reference/roxieData16T.xml')

        # assert
        with self.assertRaises(IndexError):
            RoxieToolAdapter.parse_roxie_xml(roxie_data_xml_path)

    @patch("plotly.graph_objects.Figure.show")
    def test_parse_roxie_xml_16T_raw(self, mock_show=None):
        # arrange
        roxie_data_xml_path = create_resources_file_path('resources/geometry/roxie/16T/roxieData.xml')
        roxie_data_formatted_xml_path = create_resources_file_path(
            'resources/geometry/roxie/16T/roxieData_formatted.xml')

        # act
        RoxieToolAdapter.correct_xml_file(roxie_data_formatted_xml_path, roxie_data_xml_path)

        strand_data = RoxieToolAdapter.parse_roxie_xml(roxie_data_formatted_xml_path)
        RoxieToolAdapter.plotly_results(strand_data, xlim=(0, 85), ylim=(0, 85))

        # assert
        if mock_show is not None:
            mock_show.assert_called()



from magnumapi.geometry.definitions.LayerDefinition import LayerDefinition


def test_layer_definition():
    # arrange
    layer_def = LayerDefinition(no=1, symm=1, typexy=1, blocks=[1, 2, 3, 4])

    # act
    dct = layer_def.to_roxie_dict()

    # assert
    assert dict(no=1, symm=1, typexy=1, blocks='1 2 3 4') == dct



from magnumapi.cadata.CableDatabase import CableDatabase
from magnumapi.geometry.GeometryFactory import GeometryFactory
from tests.resource_files import create_resources_file_path


def test_init_slotted_geometry():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs_slotted.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_slotted_with_json(json_path, cadata)

    # assert
    assert geometry.blocks[0].block_def.radius == 26.0
    assert geometry.blocks[4].block_def.radius == 40.35
    assert geometry.blocks[7].block_def.radius == 54.699999999999996
    assert geometry.blocks[10].block_def.radius == 69.5


def test_init_slotted_relative_geometry():
    # arrange
    json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_rel_slotted.json')
    cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_path)

    # act
    geometry = GeometryFactory.init_slotted_with_json(json_path, cadata)

    # assert
    assert geometry.blocks[0].block_def.radius == 26.0
    assert geometry.blocks[4].block_def.radius == 40.35
    assert geometry.blocks[7].block_def.radius == 54.699999999999996
    assert geometry.blocks[10].block_def.radius == 69.5



import os

import pytest

from magnumapi.tool_adapters.ansys.AnsysToolAdapter import TerminalAnsysToolAdapter

config = {"input_folder_rel_dir": "input\\Mechanic_Plate",
          "input_file": "15T_mech.inp",
          "output_file": "vallone.out",
          "model_file": "Model.inp",
          "rst_file": "15T_2d_mech.rst",
          "out_temp_file": "out_tmp_mech.txt",
          "license_type": "-aa_r",
          "executable_path": "C:\\Program Files\\ANSYS Inc\\v202\\ansys\\bin\\winx64\\ANSYS202.exe"}


def test_constructor():
    # arrange

    # act
    ansys = TerminalAnsysToolAdapter(**config)

    # assert
    assert os.getcwd() == ansys.root_dir
    assert "input\\Mechanic_Plate" == ansys.input_folder_rel_dir
    assert "15T_mech.inp" == ansys.input_file
    assert "vallone.out" == ansys.output_file
    assert "Model.inp" == ansys.model_file
    assert "15T_2d_mech.rst" == ansys.rst_file
    assert "out_tmp_mech.txt" == ansys.out_temp_file
    assert "-aa_r" == ansys.license_type
    assert "C:\\Program Files\\ANSYS Inc\\v202\\ansys\\bin\\winx64\\ANSYS202.exe" == ansys.executable_path



import unittest

import magnumapi.tool_adapters.roxie.RoxieAPI as RoxieAPI
from tests.resource_files import create_resources_file_path


class TestRoxieAPI(unittest.TestCase):
    def test_find_index_start_and_length_bottom_header_table(self):
        # arrange
        keyword = 'CABLE'

        # act
        # assert
        with self.assertRaises(Exception) as context:
            RoxieAPI.find_index_start_and_length_bottom_header_table('', keyword)

        self.assertTrue('Not found start index and length for keyword CABLE' in str(context.exception))

    def test_read_nested_bottom_header_table(self):
        # arrange
        data_path = create_resources_file_path('resources/geometry/roxie/16T/16T_22b-37-optd7f8_gx.data')

        # act
        layer_defs_df = RoxieAPI.read_nested_bottom_header_table(data_path, keyword='LAYER')


if __name__ == '__main__':
    unittest.main()



from unittest import TestCase

import pandas as pd

from magnumapi.cadata.CableDatabase import CableDatabase
from tests.resource_files import create_resources_file_path, read_csv_as_pd


class TestCableDatabase(TestCase):
    cadata_file_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
    cadata = CableDatabase.read_cadata(cadata_file_path)

    def test_write_json(self):
        json_file_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.json')
        self.cadata.write_json(json_file_path)

    def test_write_cadata(self):
        cadata_file_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2_written.cadata')
        self.cadata.write_cadata(cadata_file_path)

    def test_read_json(self):
        json_file_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.json')
        CableDatabase.read_json(json_file_path)

    def test_get_insulation_dataframe(self):
        # arrange

        # act
        insul_df = self.cadata.get_insul_df()

        # assert
        insul_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/insul.csv')
        pd.testing.assert_frame_equal(insul_df, insul_ref_df)

    def test_get_remfit_dataframe(self):
        # arrange

        # act
        remfit_df = self.cadata.get_remfit_df()

        # assert
        remfit_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/remfit.csv')
        pd.testing.assert_frame_equal(remfit_df, remfit_ref_df)

    def test_get_filament_df(self):
        # arrange

        # act
        filament_df = self.cadata.get_filament_df()

        # assert
        filament_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/filament.csv')
        pd.testing.assert_frame_equal(filament_df, filament_ref_df)

    def test_get_strand_df(self):
        # arrange

        # act
        strand_df = self.cadata.get_strand_df()

        # assert
        strand_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/strand.csv')
        pd.testing.assert_frame_equal(strand_df, strand_ref_df)

    def test_get_transient_df(self):
        # arrange

        # act
        transient_df = self.cadata.get_transient_df()

        # assert
        transient_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/transient.csv')
        pd.testing.assert_frame_equal(transient_df, transient_ref_df)

    def test_get_quench_df(self):
        # arrange

        # act
        quench_df = self.cadata.get_quench_df()

        # assert
        quench_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/quench.csv')
        pd.testing.assert_frame_equal(quench_df, quench_ref_df)

    def test_get_cable_df(self):
        # arrange

        # act
        cable_df = self.cadata.get_cable_df()

        # assert
        cable_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/cable.csv')
        cable_ref_df = cable_ref_df.fillna('')
        pd.testing.assert_frame_equal(cable_df, cable_ref_df)

    def test_get_conductor_df(self):
        # arrange

        # act
        conductor_df = self.cadata.get_conductor_df()

        # assert
        conductor_ref_df = read_csv_as_pd('resources/geometry/roxie/16T/cadata/conductor.csv')
        conductor_ref_df = conductor_ref_df.fillna('')
        pd.testing.assert_frame_equal(conductor_df, conductor_ref_df)

    def test_get_insul_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        insul_definition = self.cadata.get_insul_definition(condname)

        # assert
        self.assertEqual('16TINS', insul_definition.name)
        self.assertAlmostEqual(0.075, insul_definition.thickness, places=2)
        self.assertAlmostEqual(0.075, insul_definition.width, places=2)
        self.assertEqual('EUROCIRCOL                    ', insul_definition.comment)

    def test_get_remfit_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        remfit_definition = self.cadata.get_remfit_definition(condname)

        # assert
        self.assertEqual('LASAEU', remfit_definition.name)
        self.assertAlmostEqual(2.67845e+11, remfit_definition.c1, places=5)
        self.assertAlmostEqual(2.93800e+01, remfit_definition.c2, places=5)
        self.assertAlmostEqual(1.60000e+01, remfit_definition.c3, places=5)
        self.assertAlmostEqual(9.60000e-01, remfit_definition.c4, places=5)
        self.assertAlmostEqual(1.52000e+00, remfit_definition.c5, places=5)
        self.assertAlmostEqual(5.00000e-01, remfit_definition.c6, places=5)
        self.assertAlmostEqual(2.00000e+00, remfit_definition.c7, places=5)
        self.assertAlmostEqual(0.00000e+00, remfit_definition.c8, places=5)
        self.assertAlmostEqual(0.00000e+00, remfit_definition.c9, places=5)
        self.assertAlmostEqual(0.00000e+00, remfit_definition.c10, places=5)
        self.assertAlmostEqual(0.00000e+00, remfit_definition.c11, places=5)

        self.assertEqual(11, remfit_definition.type)
        self.assertEqual('                              ', remfit_definition.comment)

    def test_get_filament_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        filament_definition = self.cadata.get_filament_definition(condname)

        # assert
        self.assertEqual('NB3SN', filament_definition.name)
        self.assertAlmostEqual(0, filament_definition.d_fil_in, places=1)
        self.assertAlmostEqual(50.0, filament_definition.d_fil_out, places=1)
        self.assertEqual('LASAEU', filament_definition.fit_j_c)
        self.assertEqual('LASAEU', filament_definition.fit_perp)
        self.assertEqual('NB2SN TWENTE                  ', filament_definition.comment)

    def test_get_strand_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        strand_definition = self.cadata.get_strand_definition(condname)

        # assert

        self.assertAlmostEqual(14.0, strand_definition.b_ref, places=1)
        self.assertEqual('EUROCIRCOL outer              ', strand_definition.comment)
        self.assertAlmostEqual(2359.0, strand_definition.j_c_at_b_ref_t_ref, places=1)
        self.assertEqual('STR07', strand_definition.name)
        self.assertEqual(100, strand_definition.rrr)
        self.assertAlmostEqual(4.2, strand_definition.temp_ref, places=1)
        self.assertAlmostEqual(2.2, strand_definition.f_cu_nocu, places=1)
        self.assertAlmostEqual(493.3, strand_definition.dj_c_over_db, places=1)
        self.assertAlmostEqual(0.7, strand_definition.d_strand, places=1)

    def test_get_transient_definition_for_condname_default(self):
        # arrange
        condname = '16TOL8'

        # act
        transient_definition = self.cadata.get_transient_definition(condname)

        # assert
        self.assertEqual('', transient_definition.comment)
        self.assertEqual('', transient_definition.name)
        self.assertAlmostEqual(0.0, transient_definition.r_a, places=1)
        self.assertAlmostEqual(0.0, transient_definition.r_c, places=1)
        self.assertAlmostEqual(0.0, transient_definition.res_0, places=1)
        self.assertAlmostEqual(0.0, transient_definition.dres_over_db, places=1)
        self.assertAlmostEqual(0.0, transient_definition.l_fil_tp, places=1)
        self.assertAlmostEqual(0.0, transient_definition.f_strand_fill, places=1)

    def test_get_transient_definition_for_condname_nondefault(self):
        # arrange
        condname = 'YELLONIN'

        # act
        transient_definition = self.cadata.get_transient_definition(condname)

        # assert
        self.assertEqual('LHC cable resitances          ', transient_definition.comment)
        self.assertEqual('TRANS1', transient_definition.name)
        self.assertAlmostEqual(1e-7, transient_definition.r_a, places=7)
        self.assertAlmostEqual(1e-6, transient_definition.r_c, places=7)
        self.assertAlmostEqual(0.0, transient_definition.res_0, places=1)
        self.assertAlmostEqual(0.0, transient_definition.dres_over_db, places=1)
        self.assertAlmostEqual(0.0, transient_definition.l_fil_tp, places=1)
        self.assertAlmostEqual(0.0, transient_definition.f_strand_fill, places=1)

    def test_get_quench_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        quench_definition = self.cadata.get_quench_definition(condname)

        # assert
        self.assertEqual('NIST                          ', quench_definition.comment)
        self.assertEqual('NB3SN', quench_definition.name)
        self.assertEqual(3, quench_definition.res_cu)
        self.assertEqual(3, quench_definition.cp_cu)
        self.assertEqual(3, quench_definition.k_cu)
        self.assertEqual(4, quench_definition.cp_fill)
        self.assertEqual(0, quench_definition.perc_he)
        self.assertEqual(2, quench_definition.cp_ins)
        self.assertEqual(5, quench_definition.k_ins)
        self.assertEqual(3, quench_definition.cp_sc)

    def test_get_quench_definition_for_condname_empty(self):
        # arrange
        condname = '16TIL8'

        # act
        quench_definition = self.cadata.get_quench_definition(condname)

        # assert
        self.assertEqual('', quench_definition.comment)
        self.assertEqual('', quench_definition.name)
        self.assertEqual('', quench_definition.res_cu)
        self.assertEqual('', quench_definition.cp_cu)
        self.assertEqual('', quench_definition.k_cu)
        self.assertEqual('', quench_definition.cp_fill)
        self.assertEqual('', quench_definition.perc_he)
        self.assertEqual('', quench_definition.cp_ins)
        self.assertEqual('', quench_definition.k_ins)
        self.assertEqual('', quench_definition.cp_sc)


    def test_get_cable_definition_for_condname_index_error(self):
        # arrange
        condname = 'WRONG_COND_NAME'

        # act
        # assert
        with self.assertRaises(KeyError) as context:
            self.cadata.get_cable_definition(condname)

        self.assertTrue('Conductor name WRONG_COND_NAME not present in conductor definitions.'
                        in str(context.exception))

    def test_get_cable_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        cable_definition = self.cadata.get_cable_definition(condname)

        # assert
        self.assertEqual('16TOLG8', cable_definition.name)
        self.assertAlmostEqual(13.65, cable_definition.width, places=2)
        self.assertAlmostEqual(1.204, cable_definition.thickness_i, places=2)
        self.assertAlmostEqual(1.3231, cable_definition.thickness_o, places=2)
        self.assertEqual(37, cable_definition.n_s)
        self.assertEqual(100, cable_definition.l_tp)
        self.assertEqual(0, cable_definition.f_degrad)
        self.assertEqual('EUROCIRCOL 37 STR             ', cable_definition.comment)

    def test_get_conductor_definition_for_condname(self):
        # arrange
        condname = '16TOL8'

        # act
        cond_definition = self.cadata.get_conductor_definition(condname)

        # assert
        self.assertEqual('16TOLG8', cond_definition.cable_geom)
        self.assertEqual('eurocircol                    ', cond_definition.comment)
        self.assertEqual('NB3SN', cond_definition.filament)
        self.assertEqual('16TINS', cond_definition.insulation)
        self.assertEqual('16TOL8', cond_definition.name)
        self.assertEqual('NB3SN', cond_definition.quench_mat)
        self.assertEqual('STR07', cond_definition.strand)
        self.assertAlmostEqual(1.9, cond_definition.temp_ref, places=1)
        self.assertEqual('NONE', cond_definition.transient)
        self.assertEqual(1, cond_definition.type)



import math
from dataclasses import dataclass

import numpy as np
import matplotlib.pyplot as plt
import matplotlib.patches as patches

from magnumapi.geometry.primitives.Point import Point


@dataclass(frozen=True)
class Arc:
    """ An Arc class implementing a 2D line made of two end points and center along with basic geometric transformations

   Attributes:
       p_start (Point): start point of an arc
       p_center (Point): arc center
       theta (float): angle of an arc instance expressed in degrees
    """
    p_start: Point
    p_center: Point
    theta: float

    @staticmethod
    def of_end_points_center(p_start: Point, p_end: Point, p_center=Point.of_cartesian(0.0, 0.0)) -> "Arc":
        """ Method constructing an arc from two end points and a center

        :param p_start: start point
        :param p_end: end point
        :param p_center: center point
        :return: an Arc object with initialized start point, end point and theta
        """
        theta = Arc.calculate_theta(p_start.translate(-p_center),
                                    p_end.translate(-p_center))
        return Arc(p_start, p_center, theta)

    @staticmethod
    def calculate_theta_in_rad(p_start: Point, p_end: Point) -> float:
        """ Static method calculating theta angle (in radians) of an arc from two end points

        :param p_start: start point
        :param p_end: end point
        :return: theta angle of an arc instance expressed in radians
        """
        theta = p_end.get_phi_in_rad() - p_start.get_phi_in_rad()
        return theta

    @staticmethod
    def calculate_theta(p_start: Point, p_end: Point) -> float:
        """ Static method calculating theta angle (in degrees) of an arc from two end points

        :param p_start: start point
        :param p_end: end point
        :return: theta angle of an arc instance expressed in degrees
        """
        theta = p_end.get_phi() - p_start.get_phi()
        return theta

    def plot(self, ax: plt.Axes) -> None:
        """ Method plotting an arc on a matplotlib axis

        :param ax: input matplotlib axis
        """
        start_angle_deg = self.get_start_angle()
        end_angle_deg = self.get_end_angle()
        radius = self.get_radius()
        center = (self.p_center.x, self.p_center.y)
        if start_angle_deg > end_angle_deg:
            arc = patches.Arc(center, 2 * radius, 2 * radius, 0, end_angle_deg, start_angle_deg)
        else:
            arc = patches.Arc(center, 2 * radius, 2 * radius, 0, start_angle_deg, end_angle_deg)
        ax.add_patch(arc)

    def get_radius(self) -> float:
        """ Method calculating and returning an Arc instance radius

        :return: the radius of an arc instance
        """
        return (self.p_start - self.p_center).det()

    def get_start_angle_in_rad(self) -> float:
        """ Method returning a start angle of an arc instance in radians

        :return: start angle of an arc in radians
        """
        return (self.p_start - self.p_center).get_phi_in_rad()

    def get_start_angle(self) -> float:
        """ Method returning a start angle of an arc instance in degrees

        :return: start angle of an arc in degrees
        """
        return (self.p_start - self.p_center).get_phi()

    def get_end_angle_in_rad(self) -> float:
        """ Method returning an end angle of an arc instance in radians

        :return: end angle of an arc in radians
        """
        return self.get_start_angle_in_rad() + math.radians(self.theta)

    def get_end_angle(self) -> float:
        """ Method returning an end angle of an arc instance in degrees

        :return: end angle of an arc in degrees
        """
        return self.get_start_angle() + self.theta

    def copy(self) -> "Arc":
        """ Method returning a copy of a current Arc instance

        :return: a copy of a Arc instance
        """
        return Arc(self.p_start, self.p_center, self.theta)

    @staticmethod
    def describe_ellipse_arc(x_center=0.0, y_center=0.0, a=1.0, b=1.0, start_angle=0.0, end_angle=2 * np.pi, N=100,
                             closed=False):
        """ Function creating a string description of an arc for plotly library.

        :param x_center: x-coordinate of a center of an arc
        :param y_center: y-coordinate of a center of an arc
        :param a: semi-major axes
        :param b: semi-minor axes
        :param start_angle: start angle
        :param end_angle: stop angle
        :param N: number of segments
        :param closed: True if an arc is closed, False otherwise
        :return: string representation of an arc composed of a given number of segments
        """
        t = np.linspace(start_angle, end_angle, num=N)
        x = x_center + a * np.cos(t)
        y = y_center + b * np.sin(t)
        path = f'M {x[0]}, {y[0]}'
        for k in range(1, len(t)):
            path += f'L{x[k]}, {y[k]}'
        if closed:
            path += ' Z'
        return path



from typing import Tuple, Union
from dataclasses import dataclass

import pandas as pd
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Arc import Arc
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Point import Point


@dataclass(frozen=True)
class Area:
    """ An Area class implementing a 2D area made of connected lines/arcs

   Attributes:
       lines (Tuple): tuple of Line/Arc objects
    """
    lines: Tuple[Union[Line, Arc], ...]

    @staticmethod
    def of_lines(lines: Tuple[Union[Line, Arc], ...]) -> "Area":
        """ Static method creating an Area object from a tuple of Line objects. Line objects are copied to preserve
        static behaviour and avoid side effects.

        :param lines: tuple of Line objects
        :return: An area with copy of input Line objects
        """
        return Area(lines=tuple(line.copy() for line in lines))

    @staticmethod
    def of_closed_ordered_list_of_points(points: Tuple[Point, ...]) -> "Area":
        #ToDo: add test
        #ToDo: add docstring
        lines = []
        for ix in range(len(points) - 1):
            lines.append(Line.of_end_points(points[ix], points[ix + 1]))

        lines.append(Line.of_end_points(points[-1], points[0]))

        return Area.of_lines(tuple(lines))

    def get_line(self, index: int) -> Line:
        """ Method returning a line at a given index.

        :param index: index of line to be returned
        :return: Line at the given index
        """
        if abs(index) >= len(self.lines):
            raise IndexError('The requested line index %d is out of range for a list of length %d' %
                             (index, len(self.lines)))
        return self.lines[index].copy()

    def copy(self) -> "Area":
        """ Method returning a copy of an Area instance

        :return: a new Area instance with copy of lines
        """
        return Area(lines=tuple(line.copy() for line in self.lines))

    def translate(self, vector: Point) -> "Area":
        """ Method translating an Area instance by a given vector

        :param vector: vector by which an area is translated
        :return: a translated area
        """
        return Area(lines=tuple(line.translate(vector) for line in self.lines))

    def rotate(self, angle: float) -> "Area":
        """ Method rotating area by a given angle w.r.t. origin of the coordinate system (0, 0)

        :param angle: rotation angle in degrees
        :return: rotated area
        """
        return Area(lines=tuple(line.rotate(angle) for line in self.lines))

    def plot(self, ax: plt.Axes) -> None:
        """ Method plotting an Area instance on an input axis

        :param ax: input axis for plot
        """
        for line in self.lines:
            line.plot(ax)

    def __str__(self):
        """ Method returning a string representation of an Area for a user

        :return: string representation of an Area instance
        """
        return ",".join([str(line) for line in self.lines])

    def to_df(self):
        """ Method converting a line object into a dataframe with end point coordinates of each line. Lines are
        distinguished by 1-based index.

        :return: a DataFrame with x1_i, y1_i, x2_i, y2_i columns, where i is the line index (1-based)
        """
        return pd.concat([line.to_df().add_suffix('_%d' % (i + 1)) for i, line in enumerate(self.lines)], axis=1)

    @staticmethod
    def find_trapezoid_shift_to_intercept_with_radius(radius: float, area: "Area", eps=1e-6) -> float:
        """ Method finding a shift for a trapezoidal area to get in contact with a radius.
        The method works as follows:

        - for lower (index 0) line of an area, find an intersection point with a radius
        - for upper (index 2) line of an area, find an intersection point with a radius

        - find distance between the intersection point and lower line end point closest to the radius (p2)
        - find distance between the intersection point and upper line end point closest to the radius (p1)

        - take the shortest distance

        - shift the line closes to the radius (index 3) and shift according to the bisection method between 0 and the
        shortest distance

        :param radius: radius for which the shift is calculated
        :param area: area which is shifted in order to get in contact with a radius
        :param eps: bisection method accuracy
        :return: shift distance by which a trapezoidal area needs to be shifted in order to get in contact with radius
        (up to the eps accuracy)
        """
        # find point of crossing radial segment extension with radius
        # # lower line
        line_low = area.get_line(0)

        m_low, n_low = line_low.find_line_parameters()
        p_cross_low = Line.find_intersection_with_circle(radius, m_low, n_low)

        ## upper line
        line_up = area.get_line(2)
        m_up = m_low
        n_up = line_up.p2.y - m_up * line_up.p2.x
        p_cross_up = Line.find_intersection_with_circle(radius, m_up, n_up)

        if p_cross_low is None and p_cross_up is None:
            return 0.0

        # find maximum distance from the circle
        v_dir_low = p_cross_low - line_low.p1.copy()
        det_v_dir_l = v_dir_low.det()

        if abs(det_v_dir_l) <= 1e-12:
            return 0.0

        if p_cross_up is None:
            det_v_dir_u = 0
        else:
            v_dir_up = p_cross_up - line_up.p2.copy()
            det_v_dir_u = v_dir_up.det()

        # # take the maximum
        max_v_dir = max(det_v_dir_l, det_v_dir_u)

        # shift along this axis
        v_dir_l_unit = v_dir_low.unit()

        # lower point
        p_low = p_low_transl = area.get_line(0).p1.copy()
        # upper point
        line_left = area.get_line(3)
        det_line_left = line_left.det()
        alpha_left = Line.calculate_relative_alpha_angle(line_left)
        delta_alpha = 0 if alpha_left == 90 else 180
        p_up_transl = p_low.copy().translate(Point.of_polar(det_line_left, delta_alpha + alpha_left))

        start = 0
        end = max_v_dir
        shift = eps
        while (end - start) > eps:
            shift = (end + start) / 2
            p_low_transl = p_low.translate(v_dir_l_unit * shift)
            p_up_transl = p_low_transl.copy().translate(Point.of_polar(det_line_left, delta_alpha + alpha_left))

            if Line.find_interception_parameter(radius, Line.of_end_points(p_low_transl, p_up_transl)) is not None or \
                    Line.is_line_intercepting_radius(radius, Line.of_end_points(p_low_transl, p_up_transl)):
                end = shift
            else:
                start = shift

        if Line.find_interception_parameter(radius, Line.of_end_points(p_low_transl, p_up_transl)) is not None or \
            Line.is_line_intercepting_radius(radius, Line.of_end_points(p_low_transl, p_up_transl)):
            shift = start

        return shift

    @staticmethod
    def is_outside_of_first_quadrant(area: "Area", eps=1e-30) -> bool:
        """ Method checking whether an area is fully positioned in the first quadrant.

        :param area: an Area instance for which the check is performed
        :param eps: a machine precision for comparison of 0 value
        :return: True if an Area instance is in the first quadrant, False otherwise
        """
        is_x_nonegative = all([(abs(line.p1.x) >= eps) and (abs(line.p2.x) >= eps) for line in area.lines])
        is_y_nonegative = all([(abs(line.p1.y) >= eps) and (abs(line.p2.y) >= eps) for line in area.lines])
        return not (is_x_nonegative and is_y_nonegative)

    @staticmethod
    def are_areas_overlapping(area_a: "Area", area_b: "Area") -> True:
        line_a = area_a.get_line(2)
        line_a_rev = Line.of_end_points(line_a.p2, line_a.p1)
        orient_p1 = Line.calculate_point_orientation_wrt_line(line_a_rev, area_b.get_line(0).p1)
        orient_p2 = Line.calculate_point_orientation_wrt_line(line_a_rev, area_b.get_line(0).p2)

        return orient_p1 == -1 or orient_p2 == -1



from abc import ABC, abstractmethod
import os

import papermill as pm
from papermill import PapermillExecutionError

from magnumapi.optimization import NotebookConverter
from magnumapi.optimization.OptimizationConfig import OptimizationConfig
from magnumapi.optimization.constants import ERROR_KEY
from magnumapi.commons.DirectoryManager import DirectoryManager


class ModelExecutor(ABC):
    """An abstract class for model executor used in optimization

    """

    @staticmethod
    def initialize_folder_structure(config: OptimizationConfig, optimization_folder='optimization') -> str:
        """ Method initializing the folder directory structure to store optimization results outside of the main
        directory.

        :return: a path to the output root directory where all results are stored.
        """
        DirectoryManager.create_directory_if_nonexistent(config.output_folder)
        subdirectory_name = DirectoryManager.find_output_subdirectory_name(config.output_folder)
        output_subdirectory_dir = os.path.join(config.output_folder, subdirectory_name)

        DirectoryManager.create_directory_if_nonexistent(output_subdirectory_dir)
        DirectoryManager.copy_model_input(config.input_folder, 'input', output_subdirectory_dir)
        DirectoryManager.copy_notebook_folders(config.input_folder, config.notebooks, output_subdirectory_dir)
        DirectoryManager.create_directory_if_nonexistent(os.path.join(output_subdirectory_dir, optimization_folder))
        DirectoryManager.copy_model_input(config.input_folder, optimization_folder, output_subdirectory_dir)

        logger_path = os.path.join(output_subdirectory_dir, config.logger_rel_path)
        print('The logger is saved in: %s' % logger_path)

        for config_notebook in config.notebooks:
            input_ipynb_file_path = os.path.join(output_subdirectory_dir,
                                                 config_notebook.notebook_folder,
                                                 config_notebook.notebook_name)
            output_ipynb_file_path = os.path.join(output_subdirectory_dir,
                                                  config_notebook.notebook_folder,
                                                  config_notebook.notebook_name.lower().replace('.ipynb', '_script.py'))
            NotebookConverter.convert_notebook_to_script(input_ipynb_file_path,
                                                         config_notebook.notebook_name.lower().split('.')[0],
                                                         output_ipynb_file_path)

        return output_subdirectory_dir

    @abstractmethod
    def execute(self, model_dir: str, model_name: str, parameters_dct: dict) -> dict:
        """Method executing a model and returning figures of merit.

        :param model_dir: model directory
        :param model_name: name of a model
        :param parameters_dct: a dictionary with model execution parameters and corresponding values

        :return: a dictionary with figures of merit if the computation was successful, otherwise an empty dictionary
        """
        raise NotImplementedError('This method is not implemented for this class')


class ScriptModelExecutor(ModelExecutor):
    """ An implementation of ModelExecutor abstract class for scripts

    """

    def execute(self, model_dir: str, model_name: str, parameters_dct: dict) -> dict:
        """Method calculating figures of merit with scripts. Notebooks are converted to scripts.

        :param model_dir: model directory
        :param model_name: name of a model
        :param parameters_dct: a dictionary with model execution parameters and corresponding values

        :return: a dictionary with figures of merit if the computation was successful, otherwise a dictionary with -1
        error code is returned.
        """
        script = model_name.split('.')[0].lower() + '_script'
        cwd = os.getcwd()

        os.chdir(model_dir)
        run = getattr(__import__(script), 'run_' + script)
        print('Running %s script' % script)
        try:
            fom_model = run(**parameters_dct)
        except Exception as exception:
            print(exception)
            return {ERROR_KEY: -1}
        os.chdir(cwd)

        return fom_model


class NotebookModelExecutor(ModelExecutor):
    """ An implementation of ModelExecutor abstract class for notebooks

    """
    def execute(self, model_dir: str, model_name: str, parameters_dct: dict) -> dict:
        """Method calculating figures of merit with notebooks (papermill and scrapbook packages).

        :return: a dictionary with figures of merit if the computation was successful, otherwise a dictionary with -1
        error code is returned.
        """
        notebook_path = os.path.join(model_dir, model_name)
        notebook_name_split = model_name.split('.')
        out_notebook_name = '%s_out.%s' % tuple(notebook_name_split)

        out_notebook_path = os.path.join(model_dir, out_notebook_name)

        try:
            pm.execute_notebook(notebook_path, out_notebook_path, cwd=model_dir, parameters=parameters_dct)
        except PapermillExecutionError as e:
            # on error print the message
            print(e.exec_count)
            print(e.source)
            print(e.traceback[-1])
            return {ERROR_KEY: -1}
        except Exception as exception:
            raise Exception(exception)

        # fetch figure of merit
        return sb.read_notebook(out_notebook_path).scraps['model_results'].data


class ModelExecutorFactory:
    """ A factory class returning either a script or notebook executor

    """
    @staticmethod
    def build(is_script_executed: bool) -> "ModelExecutor":
        """

        :param is_script_executed: True if notebooks are executed as script, False otherwise
        :return: ScriptModelExecutor instance if True, otherwise NotebookModelExecutor
        """
        if is_script_executed:
            return ScriptModelExecutor()
        else:
            return NotebookModelExecutor()



import math
from abc import abstractmethod
from copy import deepcopy

import pandas as pd
import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Arc import Arc
from magnumapi.geometry.primitives.Area import Area
from magnumapi.geometry.blocks.Block import Block
from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.primitives.Point import Point
from magnumapi.geometry.definitions.HomogenizedBlockDefinition import HomogenizedBlockDefinition
from magnumapi.geometry.definitions.AbsoluteCosThetaBlockDefinition import AbsoluteCosThetaBlockDefinition
from magnumapi.cadata.ConductorDefinition import ConductorDefinition
from magnumapi.geometry.definitions.CosThetaBlockDefinition import CosThetaBlockDefinition
from magnumapi.geometry.definitions.RelativeCosThetaBlockDefinition import RelativeCosThetaBlockDefinition
from magnumapi.cadata.CableDefinition import CableDefinition
from magnumapi.cadata.InsulationDefinition import InsulationDefinition
from magnumapi.cadata.StrandDefinition import StrandDefinition


class CosThetaBlock(Block):
    """ A CosThetaBlock class implementing a container with block definitions from cadata input and cos-theta geometry.
    In addition, the class stores initialized coordinates of all conductors per block of cos-theta type.
    It is a base class for absolute and relative cos-theta blocks.
    """

    def __init__(self,
                 block_def: CosThetaBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        """ Constructs a CosThetaBlock abstract class instance

        :param cable_def: a cable definition from cadata
        :param insul_def: an insulation definition from cadata
        :param strand_def: a strand definition from cadata
        :param conductor_def: a conductor definition from cadata
        """
        super().__init__(cable_def=cable_def,
                         insul_def=insul_def,
                         strand_def=strand_def,
                         conductor_def=conductor_def)
        self.block_def = block_def

    @abstractmethod
    def get_alpha(self):
        """ Abstract method for returning the inclination angle

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def get_phi(self):
        """ Abstract method for returning the positioning angle

        """
        raise NotImplementedError('This method is not implemented for this class')

    def get_radius(self):
        """ Method for returning the block radius

        """
        return self.block_def.radius

    def get_nco(self):
        """ Abstract method for returning the positioning angle

        """
        return self.block_def.nco

    def build_block(self, phi_ref=0.0, alpha_ref=0.0) -> None:
        """ Method building a cos-theta block for absolute and relative definition

        :param phi_ref: reference positioning angle from a previous block (0 by default)
        :param alpha_ref: reference inclination angle from a previous block (0 by default)
        """
        self.areas = []  # to avoid creating duplicate blocks
        phi = phi_ref + self.get_phi()
        alpha = alpha_ref + self.get_alpha()
        p_ref = Point.of_polar(self.get_radius(), phi)

        for nco_i in range(self.get_nco()):
            area = self.get_insulated_isosceles_trapezium(p_ref=p_ref, alpha_ref=alpha)
            area_shift = area.copy()
            if self.is_area_intersecting_or_within_radius(area_shift):
                alpha_top_line = Line.calculate_relative_alpha_angle(area_shift.get_line(0))
                area_shift = area.translate(Point.of_polar(2 * self.cable_def.thickness_i, alpha_top_line))

            shift = Area.find_trapezoid_shift_to_intercept_with_radius(self.get_radius(), area_shift)
            alpha_low = Line.calculate_relative_alpha_angle(area_shift.get_line(0))
            area = area_shift.copy().translate(Point.of_polar(-shift, alpha_low))

            self.areas.append(area)

            p_ref = area.get_line(2).p2.copy()
            alpha = Line.calculate_relative_alpha_angle(self.areas[-1].get_line(2))

    def is_area_intersecting_or_within_radius(self, area: Area) -> bool:
        """ Method checking whether an area of a block is either intersecting with a radius or within a radius.
        If so, then the block needs to be shifted in order to align it with the radius.

        :param area: an area for which the check is performed
        :return: True if one of the two conditions is satisfied, otherwise False
        """
        is_intercepting = Line.find_interception_parameter(self.get_radius(), area.get_line(3)) is not None
        is_within_radius = Line.is_line_intercepting_radius(radius=self.get_radius(), line=area.get_line(3))
        return is_intercepting or is_within_radius

    def get_bare_area(self, area_ins: Area) -> Area:
        """ Method returning a bare are from an insulated one

        :param area_ins: an insulated area
        :return: a bare area
        """
        line_ref = area_ins.get_line(0).copy()
        p_ref = line_ref.p1.copy()
        alpha_ref = Line.calculate_relative_alpha_angle(line_ref)
        return self.get_bare_isosceles_trapezium(p_ref=p_ref, alpha_ref=alpha_ref)

    def get_insulated_isosceles_trapezium(self, p_ref=Point.of_cartesian(0.0, 0.0), alpha_ref=0.0) -> Area:
        """ Method calculating an area as an isosceles trapezium for a cos-theta distribution
        A turn is represented as an isosceles trapezoid of a given height, inner and outer width. When inner and outer
        widths are equal, then the turn is a rectangle. The difference between the inner and outer widths gives the
        keystone angle of a cable. The inner width is in contact with the mandrel radius.

        :return: an area composed of lines
        """
        keystone_half_rad = math.atan(
            (self.cable_def.thickness_o - self.cable_def.thickness_i) / (2 * self.cable_def.width))
        keystone_half_deg = math.degrees(keystone_half_rad)

        area_bare = self.get_bare_isosceles_trapezium(p_ref=p_ref, alpha_ref=alpha_ref)

        # calculate insulation coordinates
        ins_r_projected = self.insul_def.thickness / math.cos(keystone_half_rad)
        ins_a_projected = self.insul_def.width / math.cos(keystone_half_rad)

        p1_ins_r = Point.of_polar(ins_r_projected, 180 + alpha_ref)
        p1_ins_a = Point.of_polar(ins_a_projected, -90 + keystone_half_deg + alpha_ref)
        p1_ins = area_bare.get_line(0).p1 + (p1_ins_r + p1_ins_a)

        p2_ins_r = Point.of_polar(ins_r_projected, alpha_ref)
        p2_ins_a = Point.of_polar(ins_a_projected, -90 + keystone_half_deg + alpha_ref)
        p2_ins = area_bare.get_line(1).p1 + (p2_ins_r + p2_ins_a)

        p3_ins_r = Point.of_polar(ins_r_projected, 2 * keystone_half_deg + alpha_ref)
        p3_ins_a = Point.of_polar(ins_a_projected, 90 + keystone_half_deg + alpha_ref)
        p3_ins = area_bare.get_line(2).p1 + (p3_ins_r + p3_ins_a)

        p4_ins_r = Point.of_polar(ins_r_projected, 180 + 2 * keystone_half_deg + alpha_ref)
        p4_ins_a = Point.of_polar(ins_a_projected, 90 + keystone_half_deg + alpha_ref)
        p4_ins = area_bare.get_line(3).p1 + (p4_ins_r + p4_ins_a)

        return Area.of_closed_ordered_list_of_points(points=(p1_ins, p2_ins, p3_ins, p4_ins))

    def get_bare_isosceles_trapezium(self, p_ref=Point.of_cartesian(0.0, 0.0), alpha_ref=0.0) -> Area:
        """ Method calculating a bare isosceles trapezium

        :param p_ref: a reference point for a bottom left corner of the trapezium, assuming 0 alpha angle
        (0.0 by default)
        :param alpha_ref: reference inclination angle for a trapezium from previous one (0 by default)
        :return: an area with bare isosceles trapezium
        """
        keystone_half_rad = math.atan(
            (self.cable_def.thickness_o - self.cable_def.thickness_i) / (2 * self.cable_def.width))
        keystone_half_deg = math.degrees(keystone_half_rad)
        side = self.cable_def.width / math.cos(keystone_half_rad)

        p1 = p_ref
        p2 = p1 + Point.of_polar(side, alpha_ref)
        p3 = p2.copy() + Point.of_polar(self.cable_def.thickness_o, alpha_ref + 90 + keystone_half_deg)
        p4 = p1 + Point.of_polar(self.cable_def.thickness_i, alpha_ref + 90 + keystone_half_deg)

        area = Area.of_closed_ordered_list_of_points(points=(p1, p2, p3, p4))

        ins_r_projected = self.insul_def.thickness / math.cos(keystone_half_rad)
        ins_a_projected = self.insul_def.width / math.cos(keystone_half_rad)

        p1_ins_r = Point.of_polar(ins_r_projected, 180 + alpha_ref)
        p1_ins_a = Point.of_polar(ins_a_projected, -90 + keystone_half_deg + alpha_ref)

        return area.translate(-(p1_ins_r + p1_ins_a))

    def plot_block(self, ax: plt.Axes) -> None:
        """ Method plotting insulated areas corresponding to a block

        :param ax: a matplotlib axis on which a block will be plotted
        """
        circle = plt.Circle((0, 0), self.get_radius(), color='r', fill=False)
        ax.add_patch(circle)
        for area in self.areas:
            area.plot(ax)

    def plot_bare_block(self, ax: plt.Axes) -> None:
        """ Method plotting bare areas corresponding to a block

        :param ax: a matplotlib axis on which a block will be plotted
        """
        circle = plt.Circle((0, 0), self.get_radius(), color='r', fill=False)
        ax.add_patch(circle)
        for area in self.areas:
            self.get_bare_area(area).plot(ax)

    def homogenize(self):
        if not self.areas:
            raise AttributeError('In order to extract a homogenized block definition, areas need to be initialized!')

        inner_radius = self.block_def.radius
        outer_radius = inner_radius + self.cable_def.width

        # write block corner angles
        # 4 ----- 3
        # |       |
        # |       |
        # 1-------2
        phi_0 = self.areas[0].get_line(0).p1.get_phi()
        phi_1 = self.areas[0].get_line(0).p2.get_phi()
        phi_2 = self.areas[-1].get_line(2).p1.get_phi()
        phi_3 = self.areas[-1].get_line(2).p2.get_phi()

        param_dct = dict(no=self.block_def.no, type=1, nco=self.block_def.nco, radius_inner=inner_radius,
                         radius_outer=outer_radius, phi_0=phi_0, phi_1=phi_1, phi_2=phi_2, phi_3=phi_3,
                         current=self.block_def.current, condname=self.block_def.condname,
                         n1=self.block_def.n1, n2=self.block_def.n2, imag=self.block_def.imag,
                         turn=self.block_def.turn)

        return HomogenizedCosThetaBlock(block_def=HomogenizedBlockDefinition(**param_dct),
                                        cable_def=self.cable_def, strand_def=self.strand_def,
                                        conductor_def=self.conductor_def,
                                        insul_def=self.insul_def)


class AbsoluteCosThetaBlock(CosThetaBlock):
    """ An AbsoluteCosThetaBlock class implementing a container with block definitions from cadata input and
    absolute cos-theta geometry. It relies on absolute alpha and phi block angles.
    In addition, the class stores initialized coordinates of all conductors per block of cos-theta type.

    """

    def __init__(self,
                 block_def: AbsoluteCosThetaBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        """ Constructor of AbsoluteCosThetaBlock instance
        :param block_def: an absolute cos-theta block definition with radius information
        :param cable_def: a cable definition from cadata
        :param insul_def: an insulation definition from cadata
        :param strand_def: a strand definition from cadata
        :param conductor_def: a conductor definition from cadata
        """
        super().__init__(block_def, cable_def, insul_def, strand_def, conductor_def)
        self.block_def = block_def

    def get_alpha(self):
        return self.block_def.alpha

    def get_phi(self):
        return self.block_def.phi

    def to_block_df(self):
        dct_no_areas = self.to_abs_dict()
        df = pd.DataFrame(dct_no_areas, index=[0])
        return df[['no', 'type', 'nco', 'radius', 'phi', 'alpha', 'current', 'condname', 'n1', 'n2', 'imag', 'turn']]

    def to_rel_dict(self, alpha_ref=0.0, phi_ref=0.0) -> dict:
        dct = self.to_abs_dict()
        dct['alpha_r'] = dct['alpha'] - alpha_ref
        dct['phi_r'] = dct['phi'] - phi_ref
        dct.pop('alpha')
        dct.pop('phi')
        return dct

    def to_abs_dict(self):
        return deepcopy(self.block_def.__dict__)


class RelativeCosThetaBlock(CosThetaBlock):
    """ A RelativeCosThetaBlock class implementing a container with block definitions from cadata input and
    relative cos-theta geometry. It relies on relative alpha and phi block angles.
    In addition, the class stores initialized coordinates of all conductors per block of cos-theta type.

    """

    def __init__(self,
                 block_def: RelativeCosThetaBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition):
        """ Constructor of RelativeCosThetaBlock instance

        :param block_def: an absolute cos-theta block definition with radius information
        :param cable_def: a cable definition from cadata
        :param insul_def: an insulation definition from cadata
        :param strand_def: a strand definition from cadata
        :param conductor_def: a conductor definition from cadata
        """
        super().__init__(block_def, cable_def, insul_def, strand_def, conductor_def)
        self.block_def = block_def

    def get_alpha(self):
        return self.block_def.alpha_r

    def get_phi(self):
        return self.block_def.phi_r

    def to_block_df(self):
        dct_no_areas = self.to_rel_dict()
        df = pd.DataFrame(dct_no_areas, index=[0])
        return df[['no', 'type', 'nco', 'radius', 'phi_r', 'alpha_r', 'current', 'condname', 'n1', 'n2', 'imag', 'turn']]

    def to_abs_dict(self):
        dct_no_areas = deepcopy(self.block_def.__dict__)
        line_ref = self.areas[0].get_line(0)
        dct_no_areas['alpha'] = Line.calculate_relative_alpha_angle(line_ref)
        dct_no_areas['phi'] = Line.calculate_positioning_angle(line_ref, self.block_def.radius)
        dct_no_areas.pop('alpha_r')
        dct_no_areas.pop('phi_r')
        return dct_no_areas

    def to_rel_dict(self, alpha_ref=0.0, phi_ref=0.0):
        return deepcopy(self.block_def.__dict__)


class HomogenizedCosThetaBlock(Block):
    """ A HomogenizedCosThetaBlock class implementing a container with block definitions from cadata input and
    homogenized cos-theta geometry.
    In addition, the class stores initialized coordinates of all conductors per block of homogenized cos-theta type.

    """

    def __init__(self,
                 block_def: HomogenizedBlockDefinition,
                 cable_def: CableDefinition,
                 insul_def: InsulationDefinition,
                 strand_def: StrandDefinition,
                 conductor_def: ConductorDefinition) -> None:
        super().__init__(cable_def, insul_def, strand_def, conductor_def)
        self.block_def = block_def

    def to_block_df(self):
        raise NotImplementedError('This method is not implemented for this class')

    def to_abs_dict(self):
        raise NotImplementedError('This method is not implemented for this class')

    def to_rel_dict(self, alpha_ref, phi_ref):
        raise NotImplementedError('This method is not implemented for this class')

    def plot_bare_block(self, ax: plt.Axes) -> None:
        raise NotImplementedError('This method is not implemented for this class')

    def get_bare_area(self, area_ins: Area) -> Area:
        raise NotImplementedError('This method is not implemented for this class')

    def plot_block(self, ax):
        self.areas[0].plot(ax)

    def build_block(self):
        p0 = Point.of_polar(self.block_def.radius_inner, self.block_def.phi_0)
        p1 = Point.of_polar(self.block_def.radius_outer, self.block_def.phi_1)
        p2 = Point.of_polar(self.block_def.radius_outer, self.block_def.phi_2)
        p3 = Point.of_polar(self.block_def.radius_inner, self.block_def.phi_3)
        self.areas = [Area.of_lines((Line.of_end_points(p0, p1),
                                     Arc.of_end_points_center(p1, p2),
                                     Line.of_end_points(p2, p3),
                                     Arc.of_end_points_center(p3, p0)))]

    def homogenize(self):
        return self



from unittest import TestCase
from unittest.mock import patch

import matplotlib.pyplot as plt

from magnumapi.geometry.primitives.Line import Line
from magnumapi.geometry.blocks.RectangularBlock import RectangularBlock
from magnumapi.geometry.definitions.RectangularBlockDefinition import RectangularBlockDefinition
from magnumapi.cadata.CableDatabase import CableDatabase
from tests.resource_files import create_resources_file_path


class TestRectangularBlock(TestCase):
    cadata_file_path = create_resources_file_path('resources/geometry/roxie/SMC/roxie-SMC.cadata')
    cadata = CableDatabase.read_cadata(cadata_file_path)

    # BLOCK 1
    #     1     2     1           40            0            0        10000   07NB3SN2   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 29 HFDIP             0.19         0.19 'HF Dipole FNAL                '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    #  98 07NB3SN2         14.36        1.077        1.403    40          160           10 'HF Dipole FNAL                '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 101 07NB3SN2      1 07NB3SN2    HFDIP     HFM46     HFDIP     OST                1.9 'HF Dipole FNAL                '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    def test_trapezoidal_error(self):
        # arrange
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=0.0, nco=1, condname='07NB3SN2', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        # act
        with self.assertRaises(AttributeError) as context:
            RectangularBlock(block_def=block_rel_def,
                             cable_def=self.cadata.get_cable_definition('07NB3SN2'),
                             insul_def=self.cadata.get_insul_definition('07NB3SN2'),
                             strand_def=self.cadata.get_strand_definition('07NB3SN2'),
                             conductor_def=self.cadata.get_conductor_definition('07NB3SN2'))

        # assert
        self.assertTrue('Rectangular blocks do not work with trapezoidal cables '
                        '(thickness_i = 1.0770, thickness_o = 1.4030).' in str(context.exception))

    # BLOCK 1
    #     1     2     1           40            0            0        10000  SMC11T100   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 28 TININS100          0.1          0.1 'Nb3Sn 100 um after rect.      '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    # 100 SMC11T2         14.847        1.305        1.305    40           90           10 'Guess 2 for rect. of FNAL_NC  '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 106 SMC11T100     1 SMC11T2     HFMD07B   HFM46     TININS100 OST                1.9 'SMC Rect of FNAL_NC 35 turns  '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    @patch("matplotlib.figure.Figure.show")
    def test_get_first_rect_with_insulation_0_alpha(self, mock_show=None):
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=0.0, nco=1, condname='SMC11T100', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        block = RectangularBlock(block_def=block_rel_def,
                                 cable_def=self.cadata.get_cable_definition('SMC11T100'),
                                 insul_def=self.cadata.get_insul_definition('SMC11T100'),
                                 strand_def=self.cadata.get_strand_definition('SMC11T100'),
                                 conductor_def=self.cadata.get_conductor_definition('SMC11T100'))

        block.build_block()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        block.plot_block(ax)
        fig.show()

        area = block.areas[0]

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.0000
        # y(1) = 1.5050
        # x(2) = 40.0000
        # y(2) = 0.0000
        # x(3) = 55.0470
        # y(3) = 0.0000
        # x(4) = 55.0470
        # y(4) = 1.5050
        self.assertAlmostEqual(40.0000, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(55.0470, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(1.5050, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 2.1547
        # Alph0 (1-4) = 0.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 0.0000
        self.assertAlmostEqual(2.1547, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(0.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 40.1000
        # y(1) = 1.4050
        # x(2) = 40.1000
        # y(2) = 0.1000
        # x(3) = 54.9470
        # y(3) = 0.1000
        # x(4) = 54.9470
        # y(4) = 1.4050

        area_bare = block.get_bare_area(area)

        self.assertAlmostEqual(40.1000, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.4050, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(54.9470, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(1.4050, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     2     1           40            0           45        10000  SMC11T100   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 28 TININS100          0.1          0.1 'Nb3Sn 100 um after rect.      '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    # 100 SMC11T2         14.847        1.305        1.305    40           90           10 'Guess 2 for rect. of FNAL_NC  '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 106 SMC11T100     1 SMC11T2     HFMD07B   HFM46     TININS100 OST                1.9 'SMC Rect of FNAL_NC 35 turns  '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    @patch("matplotlib.figure.Figure.show")
    def test_get_first_rect_with_insulation_45_alpha(self, mock_show=None):
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=45.0, nco=1, condname='SMC11T100', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        block = RectangularBlock(block_def=block_rel_def,
                                 cable_def=self.cadata.get_cable_definition('SMC11T100'),
                                 insul_def=self.cadata.get_insul_definition('SMC11T100'),
                                 strand_def=self.cadata.get_strand_definition('SMC11T100'),
                                 conductor_def=self.cadata.get_conductor_definition('SMC11T100'))

        block.build_block()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        block.plot_block(ax)
        fig.show()

        area = block.areas[0]

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 38.9358
        # y(1) = 1.0642
        # x(2) = 40.0000
        # y(2) = 0.0000
        # x(3) = 50.6398
        # y(3) = 10.6398
        # x(4) = 49.5756
        # y(4) = 11.7040
        self.assertAlmostEqual(38.9358, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.0642, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(50.6398, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(10.6398, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(49.5756, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(11.7040, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 1.5656
        # Alph0 (1-4) = 45.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 45.0000
        self.assertAlmostEqual(1.5656, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(45.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(45.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 39.0772
        # y(1) = 1.0642
        # x(2) = 40.0000
        # y(2) = 0.1414
        # x(3) = 50.4984
        # y(3) = 10.6398
        # x(4) = 49.5756
        # y(4) = 11.5626

        area_bare = block.get_bare_area(area)

        self.assertAlmostEqual(39.0772, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(1.0642, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1414, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(50.4984, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(10.6398, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(49.5756, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(11.5626, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     2     1           40            0           90        10000  SMC11T100   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 28 TININS100          0.1          0.1 'Nb3Sn 100 um after rect.      '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    # 100 SMC11T2         14.847        1.305        1.305    40           90           10 'Guess 2 for rect. of FNAL_NC  '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 106 SMC11T100     1 SMC11T2     HFMD07B   HFM46     TININS100 OST                1.9 'SMC Rect of FNAL_NC 35 turns  '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    @patch("matplotlib.figure.Figure.show")
    def test_get_first_rect_with_insulation_90_alpha(self, mock_show=None):
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=90.0, nco=1, condname='SMC11T100', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        block = RectangularBlock(block_def=block_rel_def,
                                 cable_def=self.cadata.get_cable_definition('SMC11T100'),
                                 insul_def=self.cadata.get_insul_definition('SMC11T100'),
                                 strand_def=self.cadata.get_strand_definition('SMC11T100'),
                                 conductor_def=self.cadata.get_conductor_definition('SMC11T100'))

        block.build_block()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)
        ax.set_xlim(0, 75)
        ax.set_ylim(0, 75)

        block.plot_block(ax)
        fig.show()

        area = block.areas[0]

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 38.4950
        # y(1) = 0.0000
        # x(2) = 40.0000
        # y(2) = 0.0000
        # x(3) = 40.0000
        # y(3) = 15.0470
        # x(4) = 38.4950
        # y(4) = 15.0470
        self.assertAlmostEqual(38.4950, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(15.0470, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(38.4950, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(15.0470, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 0.0000
        # Alph0 (1-4) = 90.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 90.0000
        self.assertAlmostEqual(0.0000, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(90.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(90.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 38.5950
        # y(1) = 0.1000
        # x(2) = 39.9000
        # y(2) = 0.1000
        # x(3) = 39.9000
        # y(3) = 14.9470
        # x(4) = 38.5950
        # y(4) = 14.9470

        area_bare = block.get_bare_area(area)

        self.assertAlmostEqual(38.5950, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(39.9000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.1000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(39.9000, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(14.9470, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(38.5950, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(14.9470, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()

    # BLOCK 1
    #     1     2     1           40            0           90        10000  SMC11T100   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 28 TININS100          0.1          0.1 'Nb3Sn 100 um after rect.      '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    # 100 SMC11T2         14.847        1.305        1.305    40           90           10 'Guess 2 for rect. of FNAL_NC  '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 106 SMC11T100     1 SMC11T2     HFMD07B   HFM46     TININS100 OST                1.9 'SMC Rect of FNAL_NC 35 turns  '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    @patch("matplotlib.figure.Figure.show")
    def test_get_first_rect_with_insulation_neg90_alpha(self, mock_show=None):
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=-90.0, nco=1, condname='SMC11T100', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        block = RectangularBlock(block_def=block_rel_def,
                                 cable_def=self.cadata.get_cable_definition('SMC11T100'),
                                 insul_def=self.cadata.get_insul_definition('SMC11T100'),
                                 strand_def=self.cadata.get_strand_definition('SMC11T100'),
                                 conductor_def=self.cadata.get_conductor_definition('SMC11T100'))

        block.build_block()

        fig, ax = plt.subplots(figsize=(15, 15))
        ax.set_aspect('equal', 'box')
        ax.grid(True)

        block.plot_block(ax)
        fig.show()

        area = block.areas[0]

        # assert positions - insulated
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 41.5050
        # y(1) = 0.0000
        # x(2) = 40.0000
        # y(2) = 0.0000
        # x(3) = 40.0000
        # y(3) = -15.0470
        # x(4) = 38.4950
        # y(4) = -15.0470
        self.assertAlmostEqual(41.5050, round(area.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(0.0000, round(area.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(40.0000, round(area.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(-15.0470, round(area.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(41.5050, round(area.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(-15.0470, round(area.get_line(2).p1.y, 4), places=4)

        # Positioning angles
        # Phi0 (1-4) = 0.0000
        # Alph0 (1-4) = 270.0000
        # Phi0 (2-3) = 0.0000
        # Alph0 (2-3) = 270.0000
        # The difference is due to orientation of lines in ROXIE and MagNum.
        self.assertAlmostEqual(0.0000, area.get_line(2).p2.get_phi(), places=4)
        self.assertAlmostEqual(90.0000, Line.calculate_relative_alpha_angle(area.get_line(2)), places=4)
        self.assertAlmostEqual(0.0000, area.get_line(0).p1.get_phi(), places=4)
        self.assertAlmostEqual(270.0000, Line.calculate_relative_alpha_angle(area.get_line(0)), places=4)

        # assert positions - bare
        # Conductor number: 1
        # Part of Block: 1
        # Corner co-ordinates:
        # x(1) = 41.4050
        # y(1) = -0.1000
        # x(2) = 40.1000
        # y(2) = -0.1000
        # x(3) = 40.1000
        # y(3) = -14.9470
        # x(4) = 41.4050
        # y(4) = -14.9470

        area_bare = block.get_bare_area(area)

        self.assertAlmostEqual(41.4050, round(area_bare.get_line(3).p1.x, 4), places=4)
        self.assertAlmostEqual(-0.1000, round(area_bare.get_line(3).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(0).p1.x, 4), places=4)
        self.assertAlmostEqual(-0.1000, round(area_bare.get_line(0).p1.y, 4), places=4)
        self.assertAlmostEqual(40.1000, round(area_bare.get_line(1).p1.x, 4), places=4)
        self.assertAlmostEqual(-14.9470, round(area_bare.get_line(1).p1.y, 4), places=4)
        self.assertAlmostEqual(41.4050, round(area_bare.get_line(2).p1.x, 4), places=4)
        self.assertAlmostEqual(-14.9470, round(area_bare.get_line(2).p1.y, 4), places=4)

        # assert
        if mock_show is not None:
            mock_show.assert_called()


    # BLOCK 1
    #     1     2     3           40            0            0        10000  SMC11T100   2  20     0            0
    #    no  type   nco       radius          phi        alpha      current   condname  n1  n2  imag         turn

    #  INSUL 1
    # 28 TININS100          0.1          0.1 'Nb3Sn 100 um after rect.      '
    # No Name            Radial       Azimut  Comment
    #
    #  CABLE 1
    # 100 SMC11T2         14.847        1.305        1.305    40           90           10 'Guess 2 for rect. of FNAL_NC  '
    #  No Name            height      width_i      width_o    ns      transp.        degrd  Comment
    #
    # CONDUCTOR 1
    # 106 SMC11T100     1 SMC11T2     HFMD07B   HFM46     TININS100 OST                1.9 'SMC Rect of FNAL_NC 35 turns  '
    #  No Name       Type CableGeom.  Strand    Filament  Insul     Trans              T_o  Comment

    @patch("matplotlib.figure.Figure.show")
    def test_homogenize(self, mock_show=None):
        block_rel_def = RectangularBlockDefinition(x=40, y=0.0, alpha=0, nco=3, condname='SMC11T100', no=1,
                                                   current=10000, type=2, n1=2, n2=20, imag=0, turn=0)

        block = RectangularBlock(block_def=block_rel_def,
                                 cable_def=self.cadata.get_cable_definition('SMC11T100'),
                                 insul_def=self.cadata.get_insul_definition('SMC11T100'),
                                 strand_def=self.cadata.get_strand_definition('SMC11T100'),
                                 conductor_def=self.cadata.get_conductor_definition('SMC11T100'))

        block.build_block()

        block_homo = block.homogenize()
        block_homo.build_block()

        assert abs(block_homo.block_def.x - 40.0) <  1e-6
        assert abs(block_homo.block_def.y - 0.0) <  1e-6
        assert abs(block_homo.block_def.x_up_r - 55.047) <  1e-6
        assert abs(block_homo.block_def.y_up_r - 4.515) <  1e-6



import unittest

import numpy as np

from magnumapi.matpro.heat_capacities import calc_cv_cu_nist, calc_cv_nb3sn_nist
from tests.resource_files import create_resources_file_path


class Test_heat_capacities(unittest.TestCase):
    def test_calc_cv_cu_nist(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 300, 100)

        # act
        cv_cu_nist = calc_cv_cu_nist(T)

        # assert
        path = create_resources_file_path('resources/matpro/heat_capacities/cv_cu_nist.txt')
        cv_cu_nist_ref = np.loadtxt(path)
        np.testing.assert_allclose(cv_cu_nist_ref, cv_cu_nist)

    def test_calc_cv_nb3sn_nist(self):
        # arrange
        T_0 = 1.9
        T = np.linspace(T_0, 300, 100)

        # act
        cv_nb3sn_nist = calc_cv_nb3sn_nist(T, B=5.0)

        # assert
        path = create_resources_file_path('resources/matpro/heat_capacities/cv_nb3sn_nist.txt')
        cv_nb3sn_nist_ref = np.loadtxt(path)
        np.testing.assert_allclose(cv_nb3sn_nist_ref, cv_nb3sn_nist)


if __name__ == '__main__':
    unittest.main()




from abc import ABC, abstractmethod
import json
import os
from shutil import copyfile
from copy import deepcopy
from typing import List, TypeVar, Tuple, Dict

import numpy as np
import pandas as pd
from IPython.display import display
from ansys.mapdl.core import LocalMapdlPool

from magnumapi.geometry.Geometry import Geometry
from magnumapi.geometry.GeometryChange import GeometryChange
from magnumapi.optimization.DesignVariable import GeneticDesignVariable
from magnumapi.optimization.Logger import Logger
from magnumapi.optimization.ModelExecutor import ModelExecutorFactory
from magnumapi.optimization.OptimizationConfig import OptimizationConfig
from magnumapi.optimization.ParallelOptimizationConfig import ParallelOptimizationConfig
from magnumapi.optimization.OptimizationNotebookConfig import OptimizationNotebookConfig
from magnumapi.optimization.constants import PENALTY, ERROR_KEY
from magnumapi.tool_adapters.ansys.AnsysInputBuilder import AnsysInputBuilder
from magnumapi.tool_adapters.ansys.AnsysToolAdapter import AnsysToolAdapter

T = TypeVar('T')


class Optimization(ABC):
    def __init__(self, n_gen: int, n_pop: int, logger_rel_path='', output_subdirectory_dir=''):
        """ Constructor of a GeneticOptimization instance

        :param output_subdirectory_dir: a root subdirectory with output results
        """
        self.n_gen = n_gen
        self.n_pop = n_pop
        self.logger = Logger(output_subdirectory_dir, logger_rel_path)
        self.output_subdirectory_dir = output_subdirectory_dir
        self.pop = []

    def optimize(self) -> None:
        """ Method executing the main optimization loop of the genetic algorithm

        """
        # initial population of random chromosomes
        self.pop = self.initialize_population()

        # enumerate generations
        for gen in range(self.n_gen):
            print('Generation:', gen)
            # decode population
            fom_dcts = []

            for index, individual in enumerate(self.pop):
                print('\tIndividual:', index)
                try:
                    self.update_model(individual)
                    fom_dct = self.calculate_figures_of_merit(individual, {})
                except AttributeError:
                    fom_dct = {objective_config.objective: float('nan') for objective_config in self.config.objectives}
                fom_dcts.append(fom_dct)

            # log current computation results
            scores = self.calculate_scores(fom_dcts)
            for individual, fom_dct, score in zip(self.pop, fom_dcts, scores):
                self.logger.append_to_logger(self.decode_individual(individual), fom_dct, score)

            # display an update
            print(sorted(scores))
            min_logger_df = self.get_min_fitness_per_generation()
            display(min_logger_df[min_logger_df.index == min_logger_df.index[-1]])

            # apply genetic operators of selection, crossover, and mutation
            self.pop = self.update_generation(self.pop, scores)
            self.logger.save_logger()

    @abstractmethod
    def initialize_population(self) -> List[List[int]]:
        """ Abstract method initializing a population of candidate solutions

        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def update_model(self, individual: List[int]) -> None:
        """ Abstract method updating model parameters

        :param individual: a list of 0s and 1s representing a model to be optimised
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def decode_individual(self, individual: List[int]) -> Dict:
        """ Method decoding a chromosome from a list of ints (0, 1) to a dictionary containing values of each
        design variable. There is a check whether the decoded value matches the input range (the resulting value may
        exceed the range due to the application of crossover and mutation operators).

        :param individual: a list of ints (0, 1) representing an individual's chromosome
        :return: a dictionary with all design variables
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def calculate_figures_of_merit(self, individual: List[int], fom_init: dict={}) -> dict:
        """ Abstract method calculating figures of merit

        :param individual: a list of 0s and 1s representing a model to be optimised
        :param fom_init: an initial value of the figure of merit dictionary
        """
        raise NotImplementedError('This method is not implemented for this class')

    @abstractmethod
    def calculate_score(self, fom_dct: Dict[str, float]) -> float:
        """ Method calculating score from a dictionary mapping objective variable to its value.
        If any of the returned values is NaN, then the penalty value is returned.

        :param fom_dct: figure of merit dictionary mapping objective variable to its value
        :return: fitness function value obtained as weighted sum
        """
        raise NotImplementedError('This method is not implemented for this class')

    def calculate_scores(self, fom_dcts: list) -> list:
        return [self.calculate_score(fom_dct) for fom_dct in fom_dcts]

    @abstractmethod
    def update_generation(self, pop: List[List[int]], scores: List[float], k_selection=3) -> List[List[int]]:
        raise NotImplementedError('This method is not implemented for this class')

    def get_logger_df(self) -> pd.DataFrame:
        return self.logger.get_logger_df()

    def get_mean_fitness_per_generation(self) -> pd.DataFrame:
        return self.logger.get_mean_fitness_per_generation(self.n_pop)

    def get_min_fitness_per_generation(self) -> pd.DataFrame:
        return self.logger.get_min_fitness_per_generation(self.n_pop)


class GeneticOptimization(Optimization, ABC):
    """ Base GeneticOptimization class implementing a basic genetic optimization algorithm.

    """

    def __init__(self,
                 config: OptimizationConfig,
                 design_variables_df: pd.DataFrame,
                 n_elite=2,
                 output_subdirectory_dir='') -> None:
        """ Constructor of a GeneticOptimization instance

        :param config: a config for the genetic optimization
        :param design_variables_df: a dataframe with ranges of design variables
        :param n_elite: the number of individuals for elitism
        :param output_subdirectory_dir: a root subdirectory with output results
        """
        super().__init__(config.n_gen, config.n_pop, config.logger_rel_path, output_subdirectory_dir)
        self.r_cross = config.r_cross
        self.r_mut = config.r_mut
        self.config = config
        self.n_elite = n_elite
        self.design_variables = GeneticOptimization.initialize_design_variables(design_variables_df)

    @staticmethod
    def initialize_design_variables(design_variables_df: pd.DataFrame) -> List[GeneticDesignVariable]:
        """ Static method initializing design variables from a DataFrame
        # todo: add description of columns in the dataframe

        :param design_variables_df: a DataFrame with design variables
        :return: a list of design variables
        """
        design_variables = []
        for index, row in design_variables_df.iterrows():
                design_variables.append(GeneticDesignVariable(**row.to_dict()))

        return design_variables

    @staticmethod
    def initialize_config(json_path: str) -> OptimizationConfig:
        return OptimizationConfig.initialize_config(json_path)

    def calculate_score(self, fom_dct: Dict[str, float]) -> float:
        """ Method calculating score from a dictionary mapping objective variable to its value.
        If any of the returned values is NaN, then the penalty value is returned.

        :param fom_dct: figure of merit dictionary mapping objective variable to its value
        :return: fitness function value obtained as weighted sum
        """
        if any([np.isnan(value) for value in fom_dct.values()]):
            return PENALTY

        score = 0
        for objective_config in self.config.objectives:
            score += objective_config.weight * (fom_dct[objective_config.objective] - objective_config.constraint)

        return score

    def update_generation(self, pop: List[List[int]], scores: List[float], k_selection=3) -> List[List[int]]:
        """ Method updating generation of individuals representing candidate solutions. The update consists of applying
        the following genetic operators:
        - application of elitism (if n_elite > 0), i.e., direct propagation of best n_elite individuals to the next
        iteration of the algorithm
        - tournament selection (with given tournament size, k_selection) of parents for mating
        - crossover of pairs of parents to produce a pair of children; crossover happens with a certain probability
        - mutation with a given probability

        :param pop: list of individuals with candidate solutions. Each candidate is represented as a bitstring
        :param scores: a list of scores corresponding to the individuals
        :param k_selection: size of the tournament for selection
        :return: a new generation of individuals after applying genetic operators to the input generation
        """
        # elitism - keep best two forward
        # # sort according to the best result
        pop_sorted = [pop_el for _, pop_el in sorted(zip(scores, pop))]
        scores_sorted = [score_el for score_el, _ in sorted(zip(scores, pop))]
        # take two from pop and scores
        elite = pop_sorted[:self.n_elite]
        # select parents
        selected = [GeneticOptimization.selection(pop_sorted, scores_sorted, k_selection)
                    for _ in range(self.n_pop - self.n_elite)]
        # create the next generation
        children = []
        for i in range(0, self.n_pop - self.n_elite, 2):
            # get selected parents in pairs
            p1, p2 = selected[i], selected[i + 1]
            # crossover and mutation
            for c in self.crossover(p1, p2):
                # mutation
                self.mutation(c)
                # store for next generation
                children.append(c)
        # replace population
        return elite + children

    @staticmethod
    def selection(pop: List[List[int]], scores: List[float], k=3) -> List[int]:
        """ Method selecting best individuals according to the fitness value. The algorithm starts with a randomly
        selected individual. Then, it selects k random indices. For each of the indices it updates the initially
        selected individual by comparing the fitness with k selected individuals.

        :param pop: population of individuals, a list of genomes; sorted according to the score in ascending order
        :param scores: a list of fitness function scores, one per genome; sorted according to the score in ascending
        order
        :param k: the size of the tournament selection
        :return: a selected genome
        """
        # first random selection
        selection_ix = np.random.randint(len(pop))
        for ix in np.random.randint(0, len(pop), k - 1):
            # check if better (e.g. perform a tournament)
            if scores[ix] < scores[selection_ix]:
                selection_ix = ix

        return pop[selection_ix]

    def crossover(self, parent1: List[int], parent2: List[int]) -> Tuple[List[int], List[int]]:
        """ Method performing a single cross-over operation between two parents and creates two children. The cross-over
        point is selected at random.

        :param parent1: first parent
        :param parent2: second parent
        :return: a tuple with two children
        """
        # children are copies of parents by default
        child1, child2 = parent1.copy(), parent2.copy()
        # check for recombination
        if np.random.rand() < self.r_cross:
            # select crossover point that is not on the end of the string
            idx_crossover = np.random.randint(1, len(parent1) - 2)
            # perform crossover
            child1 = parent1[:idx_crossover] + parent2[idx_crossover:]
            child2 = parent2[:idx_crossover] + parent1[idx_crossover:]
        return child1, child2

    def mutation(self, bitstring: List[int]) -> List[int]:
        """ Method performing a bit-flip mutation with a certain probability

        :param bitstring: input bitstring for which the mutation is performed
        :return: updated bitstring
        """
        output_bitstring = bitstring.copy()
        for i in range(len(bitstring)):
            # check for a mutation
            if np.random.rand() < self.r_mut:
                # flip the bit
                output_bitstring[i] = 1 - output_bitstring[i]

        return output_bitstring


class RoxieGeneticOptimization(GeneticOptimization):
    """ A RoxieGeneticOptimization class implementing a basic genetic optimization algorithm. The class contains methods
    for updating a ROXIE input according to the update from the genetic algorithm.

    """

    def __init__(self,
                 config: OptimizationConfig,
                 design_variables_df: pd.DataFrame,
                 geometry: Geometry,
                 model_input_path: str = '',
                 is_script_executed: bool = True,
                 n_elite=2,
                 output_subdirectory_dir: str = '') -> None:
        """ Constructor of a RoxieGeneticOptimization instance

        :param config: a config for the genetic optimization
        :param design_variables_df: a dataframe with ranges of design variables
        :param geometry: input block inputs to be updated; it is a list of dictionaries
        :param model_input_path: a path to a model input json file
        :param is_script_executed: True if notebooks are executed as script, False otherwise
        :param n_elite: the number of individuals for elitism
        :param output_subdirectory_dir: a root subdirectory with output results
        """
        super().__init__(config, design_variables_df, n_elite, output_subdirectory_dir)
        self.geometry = geometry
        self.model_input_path = model_input_path
        self.is_script_executed = is_script_executed

    def initialize_population(self) -> List[List[int]]:
        population = []
        for _ in range(self.n_pop):
            ind = RoxieGeneticOptimization.generate_random_chromosome_programmable_geometry(self.design_variables)
            population.append(ind)
        return population

    def initialize_population_old(self) -> List[List[int]]:
        pop = []
        for _ in range(self.n_pop):
            chromosome = RoxieGeneticOptimization.generate_random_chromosome(self.design_variables)
            pop.append(chromosome)
        return pop

    @staticmethod
    def generate_random_chromosome(design_variables: List[GeneticDesignVariable]) -> List[int]:
        """ Static method generating a random chromosome representing an individual based on input information.

        :param design_variables: a list with definition of design variables
        :return: a randomly generated chromosome as a list of ints (0, 1)
        """
        chromosome = []
        for design_variable in design_variables:
            chromosome.extend(design_variable.generate_random_gene())

        return chromosome

    @staticmethod
    def generate_random_chromosome_programmable_geometry(design_variables):
        n_layers = max([dv.layer for dv in design_variables if 'nbl_layer' in dv.variable])
        chromosome = []
        for layer_index in range(1, n_layers + 1):
            # randomly select spar thickness
            st_dv = [dv for dv in design_variables if dv.variable == 'spar_thickness' and dv.layer == layer_index][0]
            chromosome.extend(st_dv.generate_random_gene())

            # randomly select nco
            nco_layer_dv = [dv for dv in design_variables if dv.variable == 'nco_layer' and dv.layer == layer_index][0]
            nbl_layer_dv = [dv for dv in design_variables if dv.variable == 'nbl_layer' and dv.layer == layer_index][0]
            nco_layer = np.random.randint(nco_layer_dv.xl, nco_layer_dv.xu + 1)
            nbl_layer = np.random.randint(nbl_layer_dv.xl, nbl_layer_dv.xu + 1)
            partition = RoxieGeneticOptimization.generate_number_random_partition(nco_layer, nbl_layer)

            # convert partition blocks into bits
            nco_dv = [dv for dv in design_variables if dv.variable == 'nco' and dv.layer == layer_index][0]

            for partition_el in partition:
                chromosome.extend(GeneticDesignVariable.convert_int_to_gene(partition_el, nco_dv.bits))

            if len(partition) < nbl_layer_dv.xu:
                chromosome += [0] * (nbl_layer_dv.xu - len(partition)) * nco_dv.bits

            # initiate phi_r
            for phi_r_dv in [dv for dv in design_variables if dv.variable == 'phi_r' and dv.layer == layer_index]:
                chromosome.extend(phi_r_dv.generate_random_gene())

            # initiate alpha_rad_r
            for alp_r_dv in [dv for dv in design_variables if dv.variable == 'alpha_rad_r' and dv.layer == layer_index]:
                chromosome.extend(alp_r_dv.generate_random_gene())

        return chromosome

    @staticmethod
    def generate_number_random_partition(number, n_slices):
        random_in_range = [np.random.randint(1, number) for _ in range(n_slices - 1)]
        random_in_range = [0] + sorted(random_in_range) + [number]
        partition = [t - s for s, t in zip(random_in_range, random_in_range[1:])]
        return sorted(partition, reverse=True)

    def update_model(self, individual: List[int]) -> None:
        # decoded_chromosome: a chromosome decoded into a dictionary with keys corresponding to parameter names
        # (variable and block index) mapping to parameter values
        decoded_chromosome = self.decode_chromosome_programmable_geometry(individual)
        updated_block_inputs = self.update_model_parameters_programmable_geometry(decoded_chromosome).to_dict()

        # write updated input to json
        with open(self.model_input_path, 'w') as file:
            json.dump(updated_block_inputs, file, sort_keys=True, indent=4)

    def decode_individual(self, individual: List[int]) -> Dict:
        return self.decode_chromosome_programmable_geometry(individual)

    def decode_individual_old(self, individual: List[int]) -> Dict:
        block_variable_value = {}
        for index, design_variable in enumerate(self.design_variables):
            # extract gene
            sum_bits = sum([dv.bits for dv in self.design_variables[:index]])
            gene = individual[sum_bits: sum_bits + design_variable.bits]
            value = design_variable.convert_gene_to_value(gene)
            variable_names = design_variable.get_variable_names()

            # in case a variable_name is a list created from a range of blocks
            block_variable_value = {**block_variable_value, **dict.fromkeys(variable_names, value)}

        return block_variable_value

    def decode_chromosome_programmable_geometry(self, individual: List[int]):
        decoded = {}
        # remove nbl_layer and nco_layer from the list of design variables
        index_start = 0
        n_layers = max([dv.layer for dv in self.design_variables if 'nbl_layer' in dv.variable])
        for layer_index in range(1, n_layers + 1):
            nbl_layer_dv = [dv for dv in self.design_variables
                            if dv.variable == 'nbl_layer' and dv.layer == layer_index][0]
            # decode spar thickness
            st_dv = [dv for dv in self.design_variables
                     if dv.variable == 'spar_thickness' and dv.layer == layer_index][0]
            index_end = index_start + st_dv.bits
            value = st_dv.convert_gene_to_value(individual[index_start:index_end])
            index_start = index_end
            variable_names = st_dv.get_variable_names()
            decoded = {**decoded, **dict.fromkeys(variable_names, value)}
            # decode nco
            nco_dv = [dv for dv in self.design_variables
                      if dv.variable == 'nco' and dv.layer == layer_index][0]
            for block_index in range(nbl_layer_dv.xu):
                index_end = index_start + nco_dv.bits
                value = GeneticDesignVariable.convert_gene_to_int(individual[index_start:index_end])
                index_start = index_end
                variable_name = 'nco:%d:%d' % (layer_index, block_index + 1)
                decoded[variable_name] = value

            # decode phi_r
            for phi_r_dv in [dv for dv in self.design_variables
                             if dv.variable == 'phi_r' and dv.layer == layer_index]:
                index_end = index_start + phi_r_dv.bits
                value = phi_r_dv.convert_gene_to_value(individual[index_start:index_end])
                index_start = index_end
                variable_names = phi_r_dv.get_variable_names()
                decoded = {**decoded, **dict.fromkeys(variable_names, value)}

            # decode alpha_rad_r
            for alp_r_dv in [dv for dv in self.design_variables
                             if dv.variable == 'alpha_rad_r' and dv.layer == layer_index]:
                index_end = index_start + alp_r_dv.bits
                value = alp_r_dv.convert_gene_to_value(individual[index_start:index_end])
                index_start = index_end
                variable_names = alp_r_dv.get_variable_names()
                decoded = {**decoded, **dict.fromkeys(variable_names, value)}

        return decoded

    def update_model_parameters(self, decoded_chromosome: Dict) -> Geometry:
        """ Method updating model parameters given as ROXIE block definitions. Typically, it is a relative block
        definition.

        :param decoded_chromosome: a dictionary with keys containing a parameter name and block separated with a colon
        :return: an updated list of block definitions
        """
        geometry_model = deepcopy(self.geometry)
        for block_variable, value in decoded_chromosome.items():
            if ':-:' in block_variable:
                variable, layer_index, block_index = block_variable.split(':')
                block_index = int(block_index)
                if variable == 'current':
                    geometry_model.blocks[block_index - 1].block_def.current = value
            elif ':' in block_variable:
                variable, layer_index, block_layer_index = block_variable.split(':')
                layer_index = int(layer_index)
                block_layer_index = int(block_layer_index)

                no_block = geometry_model.layer_defs[layer_index - 1].blocks[block_layer_index - 1]
                block_index = geometry_model.get_index_in_blocks_for_layer_block_index(no_block)
                if variable == 'nco':
                    geometry_model.blocks[block_index].block_def.nco = value
                elif variable == 'alpha':
                    geometry_model.blocks[block_index].block_def.alpha = value
                elif variable == 'alpha_r':
                    geometry_model.blocks[block_index].block_def.alpha_r = value
                elif variable == 'phi':
                    geometry_model.blocks[block_index].block_def.phi = value
                elif variable == 'phi_r':
                    geometry_model.blocks[block_index].block_def.phi_r = value

        # Remove empty blocks
        return GeometryChange.update_layer_indexing(geometry_model)

    def update_model_parameters_targeted(self, decoded_chromosome: Dict) -> Geometry:
        geometry_rel = deepcopy(self.geometry)

        # update phi_r
        geometry_rel = GeometryChange.update_phi_r(geometry_rel, decoded_chromosome)

        # update nco_r
        geometry_rel = GeometryChange.update_nco_r(geometry_rel, decoded_chromosome)

        # extract absolute geometry to correct radiality
        geometry_abs = geometry_rel.to_abs_geometry()

        # correct block radiality
        geometry_abs = GeometryChange.calculate_radial_alpha(geometry_abs)

        # extract relative geometry
        geometry_rel = geometry_abs.to_rel_geometry()

        # Add more modifications here

        # update alpha_rad_r
        return GeometryChange.update_alpha_radial(geometry_rel, decoded_chromosome)

    def update_model_parameters_programmable_geometry(self, decoded_chromosome: Dict) -> Geometry:
        # update spar thickness
        geometry_rel = GeometryChange.update_spar_thickness(self.geometry, decoded_chromosome)

        # update nco varying blocks
        geometry_rel = GeometryChange.update_nco_varying_blocks(geometry_rel, decoded_chromosome)

        # update phi_r
        geometry_rel = GeometryChange.update_phi_r(geometry_rel, decoded_chromosome)

        # limit minimum phi_r
        geometry_rel = GeometryChange.limit_minimum_phi_r(geometry_rel, min_length=1)

        # calculate radiality
        geometry_abs = geometry_rel.to_abs_geometry()
        geometry_abs = GeometryChange.calculate_radial_alpha(geometry_abs)

        # update alpha_phi_rad
        geometry_rel = geometry_abs.to_rel_geometry()
        geometry_rel = GeometryChange.update_alpha_radial(geometry_rel, decoded_chromosome)

        return geometry_rel

    def calculate_figures_of_merit(self, individual: List[int], fom_dct_init: dict=None) -> dict:
        """Method calculating figures of merit with scripts or notebooks (`papermill` and scrapbook packages).

        :param individual: a list of 0s and 1s representing an individual to be executed
        :return: a dictionary with figures of merit if the computation was successful, otherwise an dictionary with NaNs
        """
        # ToDo: Move to ModelExecutor
        # Retrieve global parameters
        decoded_chromosome = self.decode_individual(individual)
        global_design_variables = {key: value for (key, value) in decoded_chromosome.items() if ':' not in key}

        fom_dct = fom_dct_init

        for notebook_config in self.config.notebooks:
            notebook_folder = notebook_config.notebook_folder
            notebook_name = notebook_config.notebook_name
            notebook_dir = os.path.join(self.output_subdirectory_dir, notebook_folder)

            # copy artefacts
            for dest, source in notebook_config.input_artefacts.items():
                copyfile(os.path.join(self.output_subdirectory_dir, source),
                         os.path.join(notebook_dir, dest))

            # set parameters
            parameters_dct = RoxieGeneticOptimization.merge_input_parameters(fom_dct,
                                                                             notebook_config,
                                                                             global_design_variables)

            # execute model
            fom_model = ModelExecutorFactory.build(self.is_script_executed). \
                execute(notebook_dir, notebook_name, parameters_dct)

            # if the error key is present in an output dictionary, an dictionary with NaNs is returned and loop is ended
            if ERROR_KEY in fom_model:
                return {objective_config.objective: float('nan') for objective_config in self.config.objectives}

            fom_dct = {**fom_dct, **fom_model}

        return fom_dct

    @staticmethod
    def merge_input_parameters(fom_dct: dict,
                               notebook_config: OptimizationNotebookConfig,
                               global_design_variables: dict) -> dict:
        """ Static method merging input parameters prior to an optimization routine execution.

        :param fom_dct: a dictionary with figures of merit calculated with scripts/notebooks
        :param notebook_config: a notebook config with input parameters dictionary
        :param global_design_variables: a dictionary with global design variables
        :return:
        """
        parameters_dct = {'full_output': False}
        for dest, source in notebook_config.input_parameters.items():
            if source in fom_dct.keys():
                parameters_dct[dest] = fom_dct[source]
            elif source in global_design_variables.keys():
                parameters_dct[dest] = global_design_variables[source]
        return parameters_dct


class ParallelRoxieGeneticOptimization(RoxieGeneticOptimization):

    @staticmethod
    def initialize_config(json_path: str) -> ParallelOptimizationConfig:
        return ParallelOptimizationConfig.initialize_config(json_path)

    def initialize_optimization(self):
        root_dir = os.path.join(self.output_subdirectory_dir, self.config.ansys.root_dir)
        template_rel_dir = os.path.join(root_dir, 'template')
        template_files = ['15T_mech.template',
                          '15T_mech_post_roxie.template',
                          '15T_mech_solu.template',
                          '15T_bc.template',
                          '15T_Coil_geo.template',
                          '15T_contact_el_m.template',
                          '15T_contact_mesh.template',
                          '15T_mat_and_elem.template',
                          '15T_geometry_main.template',
                          '15T_Yoke_geo.template',
                          'CoilBlockMacro_Roxie.mac',
                          'ContactPropMacro.mac']

        input_files = ['15T_mech_%d.inp',
                       '15T_mech_post_roxie_%d.inp',
                       '15T_mech_solu_%d.inp',
                       '15T_bc_%d.inp',
                       '15T_Coil_geo_%d.inp',
                       '15T_contact_el_m_%d.inp',
                       '15T_contact_mesh_%d.inp',
                       '15T_mat_and_elem_%d.inp',
                       '15T_geometry_main_%d.inp',
                       '15T_Yoke_geo_%d.inp',
                       'CoilBlockMacro_Roxie_%d.mac',
                       'ContactPropMacro_%d.mac']

        for index in range(self.n_pop):
            for template_file, input_file in zip(template_files, input_files):
                template_path = os.path.join(template_rel_dir, template_file)
                AnsysInputBuilder.update_input_template(template_path, index, root_dir, input_file)

    def optimize(self):
        # initial population of random chromosomes
        self.pop = self.initialize_population()

        self.initialize_optimization()

        # enumerate generations
        for gen in range(self.n_gen):
            print('Generation:', gen)

            # decode population
            fom_dcts = []
            for index, individual in enumerate(self.pop):
                print('\tIndividual:', index)
                try:
                    self.update_model(individual)
                    fom_dct = self.calculate_figures_of_merit(individual, {"index": index})
                except AttributeError:
                    fom_dct = {objective_config.objective: float('nan') for objective_config in self.config.objectives}

                fom_dcts.append(fom_dct)

            # execute ANSYS in parallel
            root_dir = os.path.join(self.output_subdirectory_dir, self.config.ansys.root_dir)
            AnsysToolAdapter.remove_ansys_output_files(root_dir, self.n_pop)

            indices = self.extract_calculated_indices(fom_dcts)
            print('indices to process: ', indices)
            pool = LocalMapdlPool(self.config.ansys.n_parallel_runners,
                                  nproc=self.config.ansys.n_proc,
                                  run_location=self.config.ansys.run_dir)
            execute_ansys_in_parallel(pool, root_dir, indices)
            pool.exit()

            # read figures of merit
            ansys_fom_dcts = AnsysToolAdapter.read_multiple_ansys_figures_of_merit(root_dir, self.n_pop)

            # merge list of figures of merit
            for i in range(self.n_pop):
                fom_dcts[i] = {**fom_dcts[i], **ansys_fom_dcts[i]}

            # calculate scores
            scores = self.calculate_scores(fom_dcts)

            # log current optimization results
            for individual, fom_dct, score in zip(self.pop, fom_dcts, scores):
                self.logger.append_to_logger(self.decode_individual(individual), fom_dct, score)

            # display an update
            print(sorted(scores))
            min_logger_df = self.get_min_fitness_per_generation()
            display(min_logger_df[min_logger_df.index == min_logger_df.index[-1]])

            # apply genetic operators of selection, crossover, and mutation
            self.pop = self.update_generation(self.pop, scores)
            self.logger.save_logger()

    def extract_calculated_indices(self, fom_dcts):
        indices = []
        for index, fom_dct in enumerate(fom_dcts):
            if all([not np.isnan(value) for value in fom_dct.values()]):
                indices.append(index)
        return indices


def execute_ansys_in_parallel(pool, root_dir, indices):
    def mapping_function(mapdl, root_dir, index):
        mapdl.clear()
        # upload input and macro files prior to model execution
        input_file = '15T_mech_%d.inp' % index
        upload_files = [input_file,
                        'Model_%d.inp' % index,
                        'forces_edit_%d.vallone' % index,
                        '15T_mech_post_roxie_%d.inp' % index,
                        '15T_mech_solu_%d.inp' % index,
                        '15T_mat_and_elem_%d.inp' % index,
                        '15T_geometry_main_%d.inp' % index,
                        '15T_Coil_geo_%d.inp' % index,
                        '15T_Yoke_geo_%d.inp' % index,
                        '15T_contact_el_m_%d.inp' % index,
                        '15T_contact_mesh_%d.inp' % index,
                        '15T_bc_%d.inp' % index,
                        'ContactPropMacro_%d.mac' % index,
                        'CoilBlockMacro_Roxie_%d.mac' % index
                        ]

        for upload_file in upload_files:
            mapdl.upload(os.path.join(root_dir, upload_file), progress_bar=False)

        # run model
        mapdl.input(os.path.join(root_dir, input_file))

        # download output file
        output_file = 'vallone_%d.out' % index
        mapdl.download(output_file, os.path.join(root_dir, output_file))
        return mapdl.parameters.routine

    inputs = [(root_dir, index) for index in indices]

    pool.map(mapping_function, inputs, progress_bar=True, wait=True)



import os

from magnumapi.commons import text_file
from magnumapi.optimization import NotebookConverter
from tests.resource_files import create_resources_file_path


def test_convert_notebook_to_script():
    # arrange
    notebook_name = 'Geometry'
    notebook_path = create_resources_file_path(os.path.join('resources/optimization', notebook_name + '.ipynb'))
    script_path = create_resources_file_path(os.path.join('resources/optimization', notebook_name + '.py'))

    # act
    NotebookConverter.convert_notebook_to_script(notebook_path, notebook_name, script_path)

    # assert
    script_ref_path = create_resources_file_path(os.path.join('resources/optimization', notebook_name + '_ref.py'))

    assert text_file.readlines(script_ref_path) == text_file.readlines(script_path)



from unittest import TestCase

from magnumapi.geometry.GeometryFactory import GeometryFactory
from magnumapi.cadata.CableDatabase import CableDatabase
from magnumapi.tool_adapters.roxie.RoxieInputBuilder import RoxieInputBuilder
from tests.resource_files import create_resources_file_path


class TestRoxieInputBuilder(TestCase):
    def test_convert_flag_dct_to_str(self):
        # arrange
        roxie_input = RoxieInputBuilder()

        # act
        flag_str = roxie_input.convert_flag_dct_to_str(roxie_input.flags)

        # assert
        flag_str_ref = '  LEND=F     LWEDG=F    LPERS=F    LQUENCH=F  LALGO=F    LMIRIRON=F \n' \
                       '  LBEMFEM=F  LPSI=F     LSOLV=F    LIRON=F    LMORPH=F   LHARD=F    \n' \
                       '  LPOSTP=F   LPEAK=T    LINMARG=F  LMARG=T    LSELF=F    LMQE=F     \n' \
                       '  LINDU=F    LEDDY=F    LSOLE=F    LFIELD3=F  LFISTR=F   LSELF3=F   \n' \
                       '  LBRICK=F   LLEAD=F    LVRML=F    LOPERA=F   LOPER20=F  LANSYS=F   \n' \
                       '  LRX2ANS=F  LANS2RX=F  LDXF=F     LMAP2D=F   LMAP3D=F   LEXPR=F    \n' \
                       '  LFIL3D=F   LFIL2D=F   LCNC=F     LANSYSCN=F LWEIRON=F  LCATIA=F   \n' \
                       '  LEXEL=F    LFORCE2D=F LAXIS=T    LIMAGX=F   LIMAGY=F   LRAEND=F   \n' \
                       '  LMARKER=F  LROLER2=F  LROLERP=F  LIMAGZZ=F  LSTEP=F    LIFF=F     \n' \
                       '  LICCA=F    LICC=F     LICCIND=F  LITERNL=F  LTOPO=F    LQUEN3=F   \n' \
                       '  LAYER=T    LEULER=T   LHEAD=T    LPLOT=T    LVERS52=T  LHARM=T    \n' \
                       '  LMATRF=F   LF3LIN=F   \n' \
                       '  /'
        self.assertEqual(flag_str_ref, flag_str)

    def test_prepare_data_file_str(self):
        self.maxDiff = None
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)
        roxie_input = RoxieInputBuilder()
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        # act
        roxie_input.cadata_path = 'C:/gitlab/magnum-api/tests/resources/geometry/roxie/16T/roxieold_2.cadata'
        roxie_input.block = geometry.to_block_df()

        # assert
        file_str = roxie_input.prepare_data_file_str()
        data_path_ref = create_resources_file_path('resources/geometry/roxie/16T/reference/16T_abs_partial_ref.data')
        with open(data_path_ref, 'r') as file:
            file_str_ref = file.read()


        #ToDo Fix failing test on CI
        # self.assertEqual(file_str_ref, file_str)

    def test_build(self):
        self.maxDiff = None
        # arrange
        json_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.json')
        cadata_path = create_resources_file_path('resources/geometry/roxie/16T/roxieold_2.cadata')
        cadata = CableDatabase.read_cadata(cadata_path)
        roxie_input = RoxieInputBuilder()
        geometry = GeometryFactory.init_with_json(json_path, cadata)

        # act
        roxie_input.cadata_path = 'C:/gitlab/magnum-api/tests/resources/geometry/roxie/16T/roxieold_2.cadata'
        roxie_input.block = geometry.to_block_df()
        roxie_input.layer = geometry.to_layer_df()

    # assert
        data_path = create_resources_file_path('resources/geometry/roxie/16T/16T_abs.data')
        template_path = create_resources_file_path('resources/geometry/roxie/16T/template.roxie')
        roxie_input.build(template_path, data_path)

        data_path_ref = create_resources_file_path('resources/geometry/roxie/16T/reference/16T_abs_ref.data')

        with open(data_path, 'r') as file:
            file_str = file.read()

        with open(data_path_ref, 'r') as file:
            file_str_ref = file.read()

        #ToDo Fix failing test on CI
        # self.assertEqual(file_str_ref, file_str)



import re
from typing import List, Tuple
from io import StringIO

import pandas as pd


def extract_objective_table(output_lines: List[str],
                            index_start_objective_table: int,
                            n_objectives: int
                            ) -> pd.DataFrame:
    """ Function extracting an objective table from ROXIE output file and converting it into a dataframe.

    :param output_lines: a list of lines from the output text file
    :param index_start_objective_table: a start index of the objective table
    :param n_objectives: the length of the objective table
    :return: dataframe with objective table
    """
    objectives_table = output_lines[index_start_objective_table: index_start_objective_table + n_objectives + 1]
    objectives_table_lines = ["\t".join(objective_table.split()) for objective_table in objectives_table]

    TESTDATA = StringIO("\n".join(objectives_table_lines))
    objective_table_df = pd.read_csv(TESTDATA, sep="\t", index_col=0)
    objective_table_df.rename(columns={'WEIGHTED': 'WEIGHTED OBJ'}, inplace=True)
    objective_table_df.drop(columns='OBJ', inplace=True)
    return objective_table_df


def find_index_start_and_length_objective_table(output_lines: List[str], table_keyword: str) -> Tuple[int, int]:
    """ Function finding a start index and lenght of an objective table in output lines from a .output ROXIE file

    :param output_lines: list of output lines
    :param table_keyword: a keyword of the objective table
    :return: a tuple with objective table start index and its length
    """
    index_objective_table = None
    n_objectives = None
    for index_line, output_line in enumerate(output_lines):
        if 'OBJECTIVE' in output_line and 'S1' in output_line:
            index_objective_table = index_line

        if table_keyword in output_line:
            matches = re.findall(r"\d+", output_line)
            if matches:
                n_objectives = int(matches[0])

    return index_objective_table, n_objectives


def convert_bottom_header_table_to_str(block_df: pd.DataFrame, keyword: str, suffix='') -> str:
    """ Function converting a dataframe with a keyword and suffix to string. The string is used to create a ROXIE .data
    input file.

    :param block_df: a dataframe with content to be converted into a bottom header table string
    :param keyword: a keyword for the table
    :param suffix: ?
    :return: a string representation of a bottom header table
    """
    keyword_and_length_str = "%s %d" % (keyword, len(block_df))
    if block_df.empty:
        return keyword_and_length_str
    else:
        block_str = block_df.to_string(index=False)
        block_str_lines = block_str.split('\n')
        block_str_lines = block_str_lines[1:] + [block_str_lines[0]]
        block_str = (suffix + '\n').join(block_str_lines)
        return "%s\n%s" % (keyword_and_length_str, block_str)


def read_bottom_header_table(roxie_file_path: str, keyword='CABLE') -> pd.DataFrame:
    """ Function reading a bottom header table from ROXIE .data, .cadata, .output files

    :param roxie_file_path: a path to a roxie file
    :param keyword: a table keyword
    :return: a dataframe with content of a table given by a keyword
    """
    with open(roxie_file_path, 'r') as text_file:
        text_file_lines = text_file.read().split('\n')

    index_start, n_lines = find_index_start_and_length_bottom_header_table(text_file_lines, keyword)
    return extract_bottom_header_table(text_file_lines, index_start, n_lines)


def find_index_start_and_length_bottom_header_table(output_lines: List[str], table_keyword: str) -> Tuple[int, int]:
    """ Function finding the start and length of a table in ROXIE .output or .cadata or .data file.
    If a table keyword can't be found, the an IndexError is thrown.

    :param output_lines: a list of lines read from a ROXIE file
    :param table_keyword: a keyword for the objective table
    :return: a tuple with an index of the start of bottom header table and its length
    """
    len_keyword = len(table_keyword)
    for index_line, output_line in enumerate(output_lines):

        if (len(output_line) >= len_keyword) and (output_line[0:len_keyword] == table_keyword):
            index_table_start = index_line
            matches = re.findall(r"\d+", output_line)
            if matches:
                n_lines_table = int(matches[0])
                return index_table_start, n_lines_table

    raise IndexError('Not found start index and length for keyword %s' % table_keyword)


def extract_bottom_header_table(text_file_lines: List[str], index_start: int, n_lines: int) -> pd.DataFrame:
    """ Function extracting a bottom header template from ROXIE .data and .cadata files. The method requires a start index
    in the table as well as the number of lines the table occupies.

    :param text_file_lines: an input list of lines from a ROXIE .data or .cadata file
    :param index_start: an index where the table starts
    :param n_lines: the number of lines occupied by the table
    :return: a dataframe with the table
    """
    value_rows = text_file_lines[index_start + 1:index_start + n_lines + 1]
    header = text_file_lines[index_start + n_lines + 1]
    if 'Comment' in header:
        columns = ','.join(header.split()[:-1])
    else:
        columns = ','.join(header.split())

    # Replace comment
    values_comma_separated = []
    comments = []
    for value_row in value_rows:
        if '\'' in value_row:
            quotation_split = value_row.split('\'')
            values_without_comment = ','.join(quotation_split[0].split())
            values_comma_separated.append(values_without_comment)
            comments.append(quotation_split[1])
        else:
            values_comma_separated.append(','.join(value_row.split()))

    TESTDATA = StringIO("\n".join([columns] + values_comma_separated))
    objective_table_df = pd.read_csv(TESTDATA, sep=",")
    if comments:
        objective_table_df['Comment'] = comments
    return objective_table_df


def read_nested_bottom_header_table(roxie_file_path: str, keyword='LAYER') -> pd.DataFrame:
    """ Function reading a bottom header table from ROXIE .data, .cadata, .output files

    :param roxie_file_path: a path to a roxie file
    :param keyword: a table keyword
    :return: a nested dataframe with content of a table given by a keyword
    """
    with open(roxie_file_path, 'r') as text_file:
        text_file_lines = text_file.read().split('\n')

    index_start, n_lines = find_index_start_and_length_bottom_header_table(text_file_lines, keyword)

    return extract_nested_bottom_header_table(text_file_lines, index_start, n_lines)


def extract_nested_bottom_header_table(text_file_lines: List[str], index_start: int, n_lines: int):
    """ Function extracting a bottom header template from ROXIE .data and .cadata files. The method requires a start index
    in the table as well as the number of lines the table occupies.

    :param text_file_lines: an input list of lines from a ROXIE .data or .cadata file
    :param index_start: an index where the table starts
    :param n_lines: the number of lines occupied by the table
    :return: a dataframe with the table
    """
    value_rows = text_file_lines[index_start + 1:index_start + n_lines + 1]
    headers = text_file_lines[index_start + n_lines + 1].split()

    header_values = []
    for value_row in value_rows:
        header_value = {}
        values = value_row.split()
        for index, header in enumerate(headers[:-1]):
            header_value[header] = int(values[index])

        # It is assumed that the remaining values for a list assigned to the last column
        # Thus it starts from len(headers) - 1 until the last but one element (the last one is \ character)
        remaining_values =values[len(headers)-1:-1]
        header_value[headers[-1]] = [int(value) for value in remaining_values]
        header_values.append(header_value)

    return pd.DataFrame(header_values)


from typing import List, Union

from magnumapi.geometry.Geometry import Geometry, RelativeCosThetaGeometry, HomogenizedCosThetaGeometry
from magnumapi.geometry.blocks.CosThetaBlock import RelativeCosThetaBlock, AbsoluteCosThetaBlock, \
    HomogenizedCosThetaBlock
from magnumapi.geometry.definitions.LayerDefinition import SlottedLayerDefinition


class SlottedGeometry(Geometry):

    def __init__(self,
                 blocks: List[Union[AbsoluteCosThetaBlock, RelativeCosThetaBlock]],
                 layer_defs: List[SlottedLayerDefinition],
                 r_aperture: float) -> None:
        """ Constructor of a SlottedGeometry class

        :param blocks: a list of instances of Block class implementations (e.g., RectangularBlock, CosThetaBlock, etc.)
        :param layer_defs: a list of layer definitions indicating symmetry type,
        and a list of blocks belonging to a layer
        :param r_aperture: aperture radius in mm
        """
        super().__init__(blocks, layer_defs)
        self.r_aperture = r_aperture

        calculate_layer_radii(self)

    def init_geometry_instance(self, blocks, geometry):
        return SlottedRelativeCosThetaGeometry(r_aperture=self.r_aperture,
                                               blocks=blocks,
                                               layer_defs=geometry.layer_defs)

    def init_geometry_dict(self, block_defs, layer_defs):
        return {'r_aperture': self.r_aperture, 'block_defs': block_defs, 'layer_defs': layer_defs}

    def init_homogenized_geometry(self, blocks, layer_defs):
        return SlottedHomogenizedCosThetaGeometry(blocks, layer_defs, self.r_aperture)


def calculate_layer_radii(geometry):
    # for each layer definition, for each block update the radius
    for layer_index, layer_def in enumerate(geometry.layer_defs):
        for block_index in layer_def.blocks:
            index_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index)
            block = geometry.blocks[index_in_blocks]
            if layer_index == 0:
                r_ref = geometry.r_aperture
            else:
                block_index_prev = geometry.layer_defs[layer_index - 1].blocks[0]
                index_prev_in_blocks = geometry.get_index_in_blocks_for_layer_block_index(block_index_prev)
                r_prev = geometry.blocks[index_prev_in_blocks].block_def.radius
                width_prev = geometry.blocks[index_prev_in_blocks].cable_def.width
                ins_width_prev = geometry.blocks[index_prev_in_blocks].insul_def.width
                r_ref = r_prev + width_prev + 2 * ins_width_prev
            block.block_def.radius = r_ref + layer_def.spar_thickness


class SlottedRelativeCosThetaGeometry(RelativeCosThetaGeometry):

    def __init__(self,
                 blocks: List[RelativeCosThetaBlock],
                 layer_defs: List[SlottedLayerDefinition],
                 r_aperture: float) -> None:
        """ Constructor of a Geometry class

        :param blocks: a list of instances of Block class implementations (e.g., RectangularBlock, CosThetaBlock, etc.)
        :param layer_defs: a list of layer definitions indicating symmetry type,
        and a list of blocks belonging to a layer
        :param r_aperture: aperture radius in mm
        """
        super().__init__(blocks, layer_defs)
        self.r_aperture = r_aperture
        calculate_layer_radii(self)

    def init_geometry_instance(self, blocks, layer_defs):
        return SlottedGeometry(r_aperture=self.r_aperture, blocks=blocks, layer_defs=layer_defs)

    def init_geometry_dict(self, block_defs, layer_defs):
        return {'r_aperture': self.r_aperture, 'block_defs': block_defs, 'layer_defs': layer_defs}


class SlottedHomogenizedCosThetaGeometry(HomogenizedCosThetaGeometry):
    """SlottedHomogenizedCosThetaGeometry class for slotted, homogenized cos-theta geometry. Creates a slotted,
    homogenized geometry from both relative and absolute cos-theta geometry definition.
    Used for creation of ANSYS models.
    """

    def __init__(self,
                 blocks: List[HomogenizedCosThetaBlock],
                 layer_defs: List[SlottedLayerDefinition],
                 r_aperture: float) -> None:
        """Constructor of HomogenizedCosThetaGeometry class
    
        :param blocks: list of HomogenizedCosThetaBlock blocks
        :param layer_defs: a list of layer definitions
        """
        super().__init__(blocks, layer_defs)
        self.r_aperture = r_aperture
        self.blocks = blocks  # Superfluous assignment to fix attribute warnings of mypy

    def to_block_df(self):
        raise NotImplementedError('This method is not implemented for this class')



import random
from unittest import TestCase
from unittest.mock import patch

import pandas as pd
import numpy as np

from magnumapi.optimization.DesignVariable import GeneticDesignVariable
from magnumapi.optimization.GeneticOptimization import GeneticOptimization
from magnumapi.optimization.constants import PENALTY
from tests.resource_files import create_resources_file_path

pop = [[1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1],
       [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0],
       [1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0],
       [0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
       [1, 0, 0, 1, 1, 1, 0, 1, 1, 1, 1],
       [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1],
       [1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0],
       [1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1],
       [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
       [1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1]]

scores = [1, 1.2, 2, 2.5, 3, 3, 4, 5, 10, 20]

json_file_path = create_resources_file_path('resources/optimization/config.json')
optimization_cfg = GeneticOptimization.initialize_config(json_file_path)
csv_path = create_resources_file_path('resources/optimization/GeneticOptimization.csv')
logger_df = pd.read_csv(csv_path, index_col=0)
logger_lst = [pd.DataFrame(record, index=[0]) for record in logger_df.to_dict('records')]


class TestGeneticOptimization(TestCase):

    @patch.multiple(GeneticOptimization, __abstractmethods__=set())
    def setUp(self) -> None:
        csv_file_path = create_resources_file_path('resources/optimization/optim_input_enlarged.csv')
        self.gen_opt = GeneticOptimization(optimization_cfg, pd.read_csv(csv_file_path))
        self.gen_opt.logger.logs = logger_lst

    def test_str_config(self):
        # arrange

        # act

        # assert
        str_repr_ref_path = create_resources_file_path('resources/optimization/config_str_representation_ref.txt')

        with open(str_repr_ref_path, 'r') as file:
            str_repr_ref = file.read()

        self.assertEqual(str_repr_ref, str(optimization_cfg))

    def test_get_logger_df(self):
        # arrange

        # act
        logger_df_act = self.gen_opt.get_logger_df()

        # assert
        pd.testing.assert_frame_equal(logger_df, logger_df_act)

    def test_get_mean_fitness_per_generation(self):
        # arrange

        # act
        logger_mean_df_act = self.gen_opt.get_mean_fitness_per_generation()

        # assert
        logger_mean_df_csv_path = create_resources_file_path('resources/optimization/GeneticOptimizationMean.csv')
        logger_mean_df = pd.read_csv(logger_mean_df_csv_path, index_col=0)
        pd.testing.assert_frame_equal(logger_mean_df, logger_mean_df_act)

    def test_get_min_fitness_per_generation(self):
        # arrange

        # act
        logger_min_df_act = self.gen_opt.get_min_fitness_per_generation()

        # assert
        logger_min_df_csv_path = create_resources_file_path('resources/optimization/GeneticOptimizationMin.csv')
        logger_mean_df = pd.read_csv(logger_min_df_csv_path, index_col=0)
        pd.testing.assert_frame_equal(logger_mean_df, logger_min_df_act)

    def test_create_new_generation_without_elitism(self):
        # arrange
        np.random.seed(0)

        # act
        self.gen_opt.n_elite = 0
        pop_act = self.gen_opt.update_generation(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], pop_act[1])
        self.assertListEqual([1, 0, 1, 0, 1, 1, 0, 0, 1, 1, 1], pop_act[2])
        self.assertListEqual([0, 1, 0, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[3])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[4])
        self.assertListEqual([1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[5])
        self.assertListEqual([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], pop_act[6])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[7])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[8])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[9])

    def test_create_new_generation_with_elitism(self):
        # arrange
        np.random.seed(0)

        # act
        self.gen_opt.n_elite = 2
        pop_act = self.gen_opt.update_generation(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[1])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], pop_act[2])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], pop_act[3])
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 1], pop_act[4])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[5])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[6])
        self.assertListEqual([1, 0, 1, 1, 1, 1, 0, 0, 1, 1, 0], pop_act[7])
        self.assertListEqual([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0], pop_act[8])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0], pop_act[9])

    def test_selection(self):
        # arrange
        np.random.seed(0)

        # act
        individual = GeneticOptimization.selection(pop, scores)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1], individual)

    def test_crossover(self):
        # arrange
        np.random.seed(0)

        p1 = [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]
        p2 = [0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 0]

        # act
        c = self.gen_opt.crossover(p1, p2)

        # assert
        self.assertListEqual([1, 0, 0, 1, 1, 1, 0, 0, 1, 1, 0], c[0])
        self.assertListEqual([0, 1, 1, 0, 1, 1, 0, 1, 1, 0, 1], c[1])

    def test_mutation(self):
        # arrange
        np.random.seed(0)
        chromosome = 10 * [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]

        # act
        mutated_chromosome = self.gen_opt.mutation(chromosome)

        # assert
        mutated_chromosome_ref = [1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0,
                                  1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0,
                                  1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0,
                                  1, 1, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1, 0, 0, 0, 1, 1, 1, 0, 1, 1, 0, 1]

        self.assertEqual(mutated_chromosome_ref, mutated_chromosome)

    def test_calculate_score_with_nan(self):
        # arrange
        fom_dct = {'b3': float('nan'), 'b5': float('nan'), 'margmi': float('nan'), 'seqv': float('nan')}

        # act
        score = self.gen_opt.calculate_score(fom_dct)

        # assert
        self.assertEqual(PENALTY, score)

    def test_calculate_score(self):
        # arrange
        fom_dct = {'b3': 10.5, 'b5': 200, 'margmi': 10, 'seqv': 111}

        # act
        score = self.gen_opt.calculate_score(fom_dct)

        # assert
        self.assertAlmostEqual(30.311, score, places=5)


